.680x0
ram_start equ 0xff0000
  TMSS_REG equ 0xa14000
  HW_VERSION equ 0xa10001
  VDP_DATA equ 0xc00000
  VDP_CTRL equ 0xc00004
  Z80_RAM equ 0xa00000
  Z80_BUSREQ equ 0xa11100
  Z80_RESET equ 0xa11200

  ;-------------------------------:
  ; exception vectors
  ;-------------------------------:

  dc.l 0xfffffe00   ; startup SP
  dc.l start        ; startup PC
  dc.l interrupt    ; bus
  dc.l interrupt    ; addr
  dc.l interrupt    ; illegal
  dc.l interrupt    ; divzero
  dc.l interrupt    ; CHK
  dc.l interrupt    ; TRAPV
  dc.l interrupt    ; priv
  dc.l interrupt    ; trace
  dc.l interrupt    ; line 1010 emulator
  dc.l interrupt    ; line 1111 emulator
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt
  dc.l interrupt    ; spurious interrupt
  dc.l interrupt    ; interrupt level 1 (lowest priority)
  dc.l extint       ; interrupt level 2 = external interrupt
  dc.l interrupt    ; interrupt level 3
  dc.l hsync        ; interrupt level 4 = H-sync interrupt
  dc.l interrupt    ; interrupt level 5
  dc.l vsync        ; interrupt level 6 = V-sync interrupt
  dc.l interrupt    ; interrupt level 7 (highest priority)
  dc.l interrupt    ; TRAP #00 exception
  dc.l interrupt    ; TRAP #01 exception
  dc.l interrupt    ; TRAP #02 exception
  dc.l interrupt    ; TRAP #03 exception
  dc.l interrupt    ; TRAP #04 exception
  dc.l interrupt    ; TRAP #05 exception
  dc.l interrupt    ; TRAP #06 exception
  dc.l interrupt    ; TRAP #07 exception
  dc.l interrupt    ; TRAP #08 exception
  dc.l interrupt    ; TRAP #09 exception
  dc.l interrupt    ; TRAP #10 exception
  dc.l interrupt    ; TRAP #11 exception
  dc.l interrupt    ; TRAP #12 exception
  dc.l interrupt    ; TRAP #13 exception
  dc.l interrupt    ; TRAP #14 exception
  dc.l interrupt    ; TRAP #15 exception
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)
  dc.l interrupt    ; Unused (reserved)

  ;-------------------------------:
  ; cartridge info header
  ;-------------------------------:

  dc.b "SEGA GENESIS    "  ; must start with "SEGA"
  dc.b "(C)---- "          ; copyright
  dc.b "2015.MAY"          ; date
  dc.b "HELLO WORLD                                     " ; cart name
  dc.b "HELLO WORLD                                     " ; cart name (alt. language)
  dc.b "GM MK-0000 -00"    ; program type / catalog number
  dc.w 0x0000                ; ROM checksum
  dc.b "J               "  ; hardware used
  dc.l 0x00000000            ; start of ROM
  dc.l 0x003fffff            ; end of ROM
  dc.l 0x00ff0000,0x00ffffff ; RAM start/end
  dc.b "            "      ; backup RAM info
  dc.b "            "      ; modem info
  dc.b "                                        " ; comment
  dc.b "JUE             "  ; regions allowed

  ;-------------------------------:
  ; exception handlers
  ;-------------------------------:

extint:
hsync:
vsync:
interrupt:
  rte

ImageJavaGrinder_image equ ram_start+0
ImageJavaGrinder_palette equ ram_start+4
ImageJavaGrinder_pattern equ ram_start+8
ImageMike_image equ ram_start+12
ImageMike_palette equ ram_start+16
ImageMike_pattern equ ram_start+20
Mandelbrots_palette equ ram_start+24
PlaySong_z80_code equ ram_start+28
PlayTitleSample_z80_code equ ram_start+32
start:
  movea.l #0xfffffe00, SP

  ; Setup registers used to talk to VDP
  movea.l #VDP_DATA, a0
  movea.l #VDP_CTRL, a1

  ; During initialization:
  ; d0 = 0
  ; d1 = data movement
  ; d2 = counter
  eor.l d0, d0

  ; Initialize TMSS
  movea.l #HW_VERSION, a2
  movea.l #TMSS_REG, a3
  move.b (a2), d1           ; A10001 test the hardware version
  andi.b #0x0f, d1
  beq.b start_init_tmss     ; branch if no TMSS
  move.l #0x53454741, (a3)  ; A14000 disable TMSS
start_init_tmss:
  move.w (a1), d1    ; C00004 read VDP status (interrupt acknowledge?)

  ; Initialize video
  movea.l #vdp_reg_init_table, a2
  move.w #0x8000, d1
  moveq #24-1, d2   ; length of video initialization block
start_video_init:
  move.b (a2)+, d1  ; get next video control byte
  move.w d1, (a1)   ; C00004 send write register command to VDP
  add.w #0x100, d1  ; point to next VDP register
  dbra d2, start_video_init  ; loop for rest of block

  ; DMA is now set up for 65535-byte fill of VRAM
  move.l #0x40000080, (a1)  ; C00004 = VRAM write to 0x0000
  move.w d0, (a0)      ; C00000 = write zero to VRAM (starts DMA fill)
  ; Wait on busy VDP
start_wait_dma:
  move.w (a1), d1      ; C00004 read VDP status
  btst.l #1, d1        ; test DMA busy flag
  bne.s start_wait_dma ; loop while DMA busy

  ; initialize CRAM
  move.l #0x81048f02, (a1) ; C00004 reg 1 = 0x04, reg 15 = 0x02: blank, auto-increment=2
  move.l #0xc0000000, (a1) ; C00004 write CRAM address 0x0000
  moveq #32-1, d2          ; loop for 32 CRAM registers
start_init_cram:
  move.l d0, (a0)          ; C00000 clear CRAM register
  dbra d2, start_init_cram

  ; Initialize VSRAM
  move.l #0x40000010, (a1) ; C00004 VSRAM write address 0x0000
  moveq #20-1, d2          ; loop for 20 VSRAM registers
start_init_vsram:
  move.l d0, (a0)          ; C00000 clear VSRAM register
  dbra d2, start_init_vsram

  ; Initialize PSG
  moveq #4-1, d2             ; loop for 4 PSG registers
  movea.l #psg_reg_init_table, a2
start_init_psg:
  move.b (a2)+, (0x0011, a0) ; C00011 copy PSG initialization commands
  dbra d2, start_init_psg

  ; Unblank display
  move.w #0x8144, (a1)   ; C00004 reg 1 = 0x44 unblank display

  ;; Set up heap and static initializers
  movea.l #ram_start+36, a5
  move.l #_ImageJavaGrinder_pattern, (ImageJavaGrinder_pattern)
  move.l #_ImageJavaGrinder_image, (ImageJavaGrinder_image)
  move.l #_ImageJavaGrinder_palette, (ImageJavaGrinder_palette)
  move.l #_ImageMike_pattern, (ImageMike_pattern)
  move.l #_ImageMike_image, (ImageMike_image)
  move.l #_ImageMike_palette, (ImageMike_palette)
  move.l #_Mandelbrots_palette, (Mandelbrots_palette)
  move.l #_PlaySong_z80_code, (PlaySong_z80_code)
  move.l #_PlayTitleSample_z80_code, (PlayTitleSample_z80_code)

main:
  link a6, #-0xc
  movea.l #ImageJavaGrinder_palette, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_palette_colors
  movea.l #ImageJavaGrinder_pattern, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_pattern_table
  movea.l #ImageJavaGrinder_image, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_image_data
  moveq #30, d0
  move.l d0, (-12,SP)
  jsr wait_I
  movea.l #PlayTitleSample_z80_code, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _load_z80
  moveq #120, d0
  move.l d0, (-12,SP)
  jsr wait_I
  move.w #0x100, (Z80_BUSREQ) ; Pause Z80
  movea.l #ImageMike_palette, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_palette_colors
  movea.l #ImageMike_pattern, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_pattern_table
  movea.l #ImageMike_image, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_image_data
  moveq #120, d0
  move.l d0, (-12,SP)
  jsr wait_I
  jsr _init_bitmap
  movea.l #PlaySong_z80_code, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _load_z80
  movea.l #Mandelbrots_palette, a2
  move.l (a2), d0
  movea.l d0, a3
  jsr _set_palette_colors
  jsr Mandelbrots_draw
  moveq #60, d0
  move.l d0, (-12,SP)
  jsr wait_I
  jsr _clear_bitmap
  jsr _load_fonts
  moveq #0, d0
  moveq #7, d1
  ; Set cursor position in VDP
  mulu.w #128, d1
  add.w d0, d1
  lsl.l #8, d1
  lsl.l #8, d1
  or.l #0x40000003, d1
  move.l d1, (a1)
  move.l #_string_67, d0
  movea.l d0, a2
  jsr _print_string
  move.l #0x40000003, (a1) ; Set cursor position in VDP
  move.l #_string_69, d0
  movea.l d0, a2
  jsr _print_string
main_110:
  move.w (a1), d0      ; C00004 read VDP status
  and.l #8, d0         ; test vertical blank flag
  cmp.l #0, d0
  beq.s main_119  ; distance=2
  bra.s main_110  ; distance=2
main_119:
  moveq #0, d0
  move.l d0 ,(-12,a6)
main_121:
  move.l (-12,a6), d0
  moveq #4, d1
  cmp.l d1, d0
  bge.s main_166  ; distance=23
  moveq #0, d0
  move.l d0 ,(-8,a6)
main_128:
  move.l (-8,a6), d0
  moveq #15, d1
  cmp.l d1, d0
  bge.s main_160  ; distance=16
  move.l #0xc0010000, (a1) ; Set CRAM write address
  move.l (-8,a6), d0
  asl.l #8, d0
  move.l (-8,a6), d1
  asl.l #4, d1
  or.l d1, d0
  move.l (-8,a6), d1
  or.l d1, d0
  move.l d0, (a0)
  ; Wait for horizontal blank
_wait_horizontal_blank_0:
  move.w (a1), d5      ; C00004 read VDP status
  btst.l #2, d5        ; test horizontal blank flag
  beq.s _wait_horizontal_blank_0
  addq.l #1, (-8,a6)
  bra.s main_128  ; distance=17
main_160:
  addq.l #1, (-12,a6)
  bra.s main_121  ; distance=24
main_166:
  ; Wait for vertical blank
_wait_vertical_blank_1:
  move.w (a1), d5      ; C00004 read VDP status
  btst.l #3, d5        ; test vertical blank flag
  beq.s _wait_vertical_blank_1
  bra.s main_110  ; distance=31

wait_I:
  link a6, #-0x8
  moveq #0, d0
  move.l d0 ,(-8,a6)
wait_I_2:
  move.l (-8,a6), d0
  move.l (-4,a6), d1
  cmp.l d1, d0
  bge.s wait_I_31  ; distance=9
wait_I_7:
  move.w (a1), d0      ; C00004 read VDP status
  and.l #8, d0         ; test vertical blank flag
  cmp.l #0, d0
  bne.s wait_I_16  ; distance=2
  bra.s wait_I_7  ; distance=2
wait_I_16:
  move.w (a1), d0      ; C00004 read VDP status
  and.l #8, d0         ; test vertical blank flag
  cmp.l #0, d0
  beq.s wait_I_25  ; distance=2
  bra.s wait_I_16  ; distance=2
wait_I_25:
  addq.l #1, (-8,a6)
  bra.s wait_I_2  ; distance=10
wait_I_31:
  unlk a6
  rts

Mandelbrots_draw:
  link a6, #-0x48
  moveq #9, d0
  move.l d0 ,(-68,a6)
  moveq #9, d0
  move.l d0 ,(-72,a6)
  move.l #0xfffffc00, d0
  move.l d0 ,(-20,a6)
  moveq #0, d0
  move.l d0 ,(-12,a6)
Mandelbrots_draw_15:
  move.l (-12,a6), d0
  move.l #0xe0, d1
  cmp.l d1, d0
  bge.w Mandelbrots_draw_165  ; distance=78
  move.l #0xfffff800, d0
  move.l d0 ,(-16,a6)
  moveq #0, d0
  move.l d0 ,(-8,a6)
Mandelbrots_draw_28:
  move.l (-8,a6), d0
  move.l #0x140, d1
  cmp.l d1, d0
  bge.w Mandelbrots_draw_152  ; distance=65
  moveq #0, d0
  move.l d0 ,(-28,a6)
  moveq #0, d0
  move.l d0 ,(-24,a6)
  moveq #0, d0
  move.l d0 ,(-48,a6)
Mandelbrots_draw_44:
  move.l (-48,a6), d0
  moveq #16, d1
  cmp.l d1, d0
  bge.w Mandelbrots_draw_123  ; distance=41
  move.l (-28,a6), d0
  move.l (-28,a6), d1
  muls.w d1, d0
  moveq #10, d1
  asr.l d1, d0
  move.l d0 ,(-40,a6)
  move.l (-24,a6), d0
  move.l (-24,a6), d1
  muls.w d1, d0
  moveq #10, d1
  asr.l d1, d0
  move.l d0 ,(-44,a6)
  move.l (-40,a6), d0
  move.l (-44,a6), d1
  add.l d1, d0
  move.l #0x1000, d1
  cmp.l d1, d0
  ble.s Mandelbrots_draw_85  ; distance=2
  bra.s Mandelbrots_draw_123  ; distance=23
Mandelbrots_draw_85:
  move.l (-40,a6), d0
  move.l (-44,a6), d1
  sub.l d1, d0
  move.l d0 ,(-32,a6)
  moveq #2, d0
  move.l (-28,a6), d1
  move.l (-24,a6), d2
  muls.w d2, d1
  moveq #10, d2
  asr.l d2, d1
  muls.w d1, d0
  move.l d0 ,(-36,a6)
  move.l (-32,a6), d0
  move.l (-16,a6), d1
  add.l d1, d0
  move.l d0 ,(-28,a6)
  move.l (-36,a6), d0
  move.l (-20,a6), d1
  add.l d1, d0
  move.l d0 ,(-24,a6)
  addq.l #1, (-48,a6)
  bra.w Mandelbrots_draw_44  ; distance=42
Mandelbrots_draw_123:
  move.l (-48,a6), d0
  moveq #16, d1
  cmp.l d1, d0
  bne.s Mandelbrots_draw_134  ; distance=3
  moveq #15, d0
  move.l d0 ,(-48,a6)
Mandelbrots_draw_134:
  move.l (-8,a6), d0
  move.l (-12,a6), d1
  move.l (-48,a6), d2
  move.l d2, d7
  move.l d1, d6
  move.l d0, d5
  jsr _plot
  move.l (-16,a6), d0
  move.l (-68,a6), d1
  add.l d1, d0
  move.l d0 ,(-16,a6)
  addq.l #1, (-8,a6)
  bra.w Mandelbrots_draw_28  ; distance=66
Mandelbrots_draw_152:
  move.l (-20,a6), d0
  move.l (-72,a6), d1
  add.l d1, d0
  move.l d0 ,(-20,a6)
  addq.l #1, (-12,a6)
  bra.w Mandelbrots_draw_15  ; distance=79
Mandelbrots_draw_165:
  unlk a6
  rts

.align 32
  dc32 2256   ; ImageJavaGrinder_pattern.length
_ImageJavaGrinder_pattern:
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0011, 0x0122, 0x3222, 0x2222
  dc32 0x0000, 0x0000, 0x0000, 0x44444400, 0x22222211, 0x22222222, 0x22222222, 0x21112222
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x23000000, 0x22000000
  dc32 0x0000, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222
  dc32 0x50000, 0x22222200, 0x22222224, 0x22222222, 0x22222222, 0x33132222, 0x1222, 0x0222
  dc32 0x0000, 0x2222, 0x2222, 0x40002222, 0x20002222, 0x21003313, 0x22000000, 0x22000000
  dc32 0x0000, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220
  dc32 0x5000, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200
  dc32 0x500000, 0x22220000, 0x22220000, 0x22220000, 0x22220000, 0x33330000, 0x0000, 0x0000
  dc32 0x0005, 0x0222, 0x0222, 0x0222, 0x0222, 0x0222, 0x0222, 0x0222
  dc32 0x0500, 0x22222222, 0x22222222, 0x22222222, 0x22222222, 0x22222222, 0x22111112, 0x22000001
  dc32 0x0000, 0x0000, 0x24000000, 0x22400000, 0x22210000, 0x22221000, 0x22222000, 0x22222300
  dc32 0x0005, 0x22222, 0x22222, 0x22222, 0x22222, 0x33333, 0x0000, 0x0000
  dc32 0x0500, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222
  dc32 0x0000, 0x22222220, 0x22222222, 0x22222222, 0x22222222, 0x33322222, 0x12222, 0x1222
  dc32 0x0000, 0x0000, 0x0000, 0x40000000, 0x21000000, 0x22000000, 0x22000000, 0x22000000
  dc32 0x0000, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002
  dc32 0x50000, 0x22220000, 0x22220000, 0x22220000, 0x22220000, 0x22220000, 0x22220000, 0x22220000
  dc32 0x2222, 0x2442, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x40004222, 0x0222, 0x0222, 0x2222, 0x42222, 0x41122222, 0x22222223, 0x22222223
  dc32 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x25000000, 0x30000000, 0x0000, 0x0000
  dc32 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222, 0x22222
  dc32 0x0222, 0x0222, 0x0222, 0x2222, 0x44442222, 0x22222223, 0x22222220, 0x22222221
  dc32 0x22000000, 0x22005505, 0x20002222, 0x20002222, 0x50002222, 0x2222, 0x2222, 0x30002222
  dc32 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220
  dc32 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200
  dc32 0x0000, 0x5500000, 0x22220000, 0x22220000, 0x22220000, 0x22220001, 0x22220002, 0x22220002
  dc32 0x0000, 0x55050, 0x222222, 0x2222222, 0x42222222, 0x22222522, 0x22223032, 0x22210001
  dc32 0x0000, 0x50000005, 0x20000022, 0x22000022, 0x22400022, 0x22210022, 0x22220022, 0x22223022
  dc32 0x0000, 0x50505050, 0x22202222, 0x22222222, 0x22222222, 0x22223222, 0x22230322, 0x22200012
  dc32 0x0000, 0x5000000, 0x22000000, 0x22000000, 0x22400000, 0x22200000, 0x22200000, 0x22200000
  dc32 0x0222, 0x0222, 0x0222, 0x0222, 0x0222, 0x0222, 0x0222, 0x0222
  dc32 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x22000000
  dc32 0x42222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200, 0x2222200
  dc32 0x0000, 0x5050, 0x22222, 0x2222222, 0x2222222, 0x12225332, 0x22220003, 0x22220000
  dc32 0x0000, 0x50000505, 0x20000222, 0x22000222, 0x22400522, 0x22210022, 0x22220022, 0x22220012
  dc32 0x0000, 0x50000005, 0x22000002, 0x22000002, 0x22000002, 0x22100012, 0x22200022, 0x22200022
  dc32 0x0000, 0x5055005, 0x22022222, 0x22022222, 0x22022222, 0x22022222, 0x22022222, 0x22022222
  dc32 0x0000, 0x0050, 0x0022, 0x2222, 0x42222, 0x22222, 0x122223, 0x222220
  dc32 0x0000, 0x50050000, 0x22220000, 0x22222000, 0x22222400, 0x32222200, 0x3222210, 0x122220
  dc32 0x0000, 0x5050, 0x22222, 0x222222, 0x4222222, 0x2222353, 0x12223000, 0x22220000
  dc32 0x0000, 0x5000000, 0x22000000, 0x22200002, 0x22240042, 0x22220022, 0x22221022, 0x22222022
  dc32 0x0000, 0x5055050, 0x22222220, 0x22222222, 0x22222222, 0x22233222, 0x22300322, 0x22130013
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x40000000, 0x21000000, 0x22000000, 0x31000000
  dc32 0x0222, 0x0222, 0x0222, 0x0222, 0x4222, 0x41142222, 0x22222223, 0x22222220
  dc32 0x22000000, 0x22055505, 0x22022222, 0x20022222, 0x20022222, 0x30022222, 0x22222, 0x22222
  dc32 0x0000, 0x5505, 0x2222, 0x2222, 0x2222, 0x2222, 0x2222, 0x2222
  dc32 0x0000, 0x55000, 0x222200, 0x222222, 0x222222, 0x222222, 0x222223, 0x222220
  dc32 0x0000, 0x5055000, 0x22222200, 0x22222200, 0x22222240, 0x52222220, 0x3222220, 0x122220
  dc32 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002
  dc32 0x22220000, 0x22220000, 0x22220000, 0x22220000, 0x22220004, 0x22220002, 0x22220002, 0x22220003
  dc32 0x0000, 0x505505, 0x2222222, 0x22222222, 0x22222222, 0x22235322, 0x22200022, 0x13101322
  dc32 0x0000, 0x55005, 0x22222, 0x20022222, 0x24052222, 0x22002222, 0x22002222, 0x22001222
  dc32 0x0000, 0x0505, 0x0222, 0x0222, 0x4222, 0x10002222, 0x20002222, 0x20002222
  dc32 0x0000, 0x50005005, 0x20002222, 0x20022222, 0x20422222, 0x31222253, 0x2222300, 0x3131313
  dc32 0x0000, 0x50500000, 0x22200000, 0x22220000, 0x22224000, 0x32222000, 0x2222000, 0x32222000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x22222, 0x22222, 0x52222, 0x2222
  dc32 0x12222222, 0x4412222, 0x4222, 0x0222, 0x0222, 0x0222, 0x40004222, 0x21112222
  dc32 0x30000000, 0x20000000, 0x25000000, 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x23000000
  dc32 0x11112222, 0x4222, 0x0422, 0x0022, 0x0022, 0x0422, 0x4222, 0x11112222
  dc32 0x23002222, 0x22002222, 0x22502222, 0x22202222, 0x22202222, 0x22202222, 0x22202222, 0x22302222
  dc32 0x22220032, 0x22220022, 0x22220022, 0x22220022, 0x22220022, 0x22220052, 0x22220002, 0x22220002
  dc32 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22240004, 0x22221012
  dc32 0x22222022, 0x22222022, 0x22222022, 0x22222022, 0x22222022, 0x22222022, 0x22225022, 0x22220022
  dc32 0x22200002, 0x22200002, 0x22200002, 0x22200002, 0x22200002, 0x22200002, 0x22200002, 0x22200002
  dc32 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22200000, 0x22200000
  dc32 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x22000000, 0x22000004, 0x22414412, 0x22222222
  dc32 0x2222203, 0x2222202, 0x2222202, 0x2222202, 0x22222202, 0x22222505, 0x22222000, 0x22222000
  dc32 0x22223333, 0x22222222, 0x22222222, 0x22222222, 0x22220000, 0x22220000, 0x22220000, 0x22221001
  dc32 0x22220002, 0x22220002, 0x22220002, 0x22220002, 0x0000, 0x0000, 0x41140000, 0x22220000
  dc32 0x22200022, 0x22250522, 0x22220222, 0x22220222, 0x22220222, 0x22224222, 0x22222222, 0x32222222
  dc32 0x21022222, 0x20022222, 0x20022222, 0x20022222, 0x22222, 0x22222, 0x22222, 0x22222
  dc32 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222220, 0x222224, 0x322222
  dc32 0x11110, 0x0000, 0x0000, 0x0000, 0x22000, 0x422440, 0x222220, 0x11222230
  dc32 0x22223333, 0x22222222, 0x22222222, 0x22222222, 0x22220000, 0x22220000, 0x22220000, 0x32221101
  dc32 0x22222022, 0x22222042, 0x22222004, 0x22222000, 0x0000, 0x0044, 0x41144022, 0x22223022
  dc32 0x22223330, 0x22222225, 0x42222222, 0x2222222, 0x22222, 0x44055222, 0x22000222, 0x22101222
  dc32 0x0000, 0x0000, 0x50000000, 0x20000000, 0x22000000, 0x22000000, 0x22000000, 0x23000000
  dc32 0x22222223, 0x41422222, 0x42222, 0x2222, 0x2222, 0x2222, 0x2222, 0x3222
  dc32 0x30022222, 0x20022222, 0x20022222, 0x20022222, 0x22022222, 0x22022222, 0x22022222, 0x22022222
  dc32 0x2222, 0x2222, 0x2222, 0x2222, 0x2222, 0x2222, 0x42222, 0x11122222
  dc32 0x22220, 0x22220, 0x22220, 0x22220, 0x22220, 0x22220, 0x22220, 0x22220
  dc32 0x0000, 0x0000, 0x0005, 0x0002, 0x0002, 0x0002, 0x0002, 0x0003
  dc32 0x0002, 0x0002, 0x55550002, 0x22220002, 0x22220002, 0x22220002, 0x22221442, 0x22222222
  dc32 0x22220000, 0x22220005, 0x22220052, 0x22220022, 0x22220022, 0x22220022, 0x22220022, 0x22230022
  dc32 0x33332222, 0x22222222, 0x22224422, 0x22220022, 0x22200022, 0x22500022, 0x22000022, 0x22100122
  dc32 0x22000222, 0x22000222, 0x22000222, 0x22000222, 0x22000022, 0x22000022, 0x22000022, 0x22000032
  dc32 0x20032221, 0x25022220, 0x22022220, 0x22022220, 0x22022200, 0x22422200, 0x22222200, 0x22222200
  dc32 0x33222, 0x522222, 0x5222224, 0x2222200, 0x2222000, 0x2222000, 0x2222000, 0x2222101
  dc32 0x22222000, 0x22222000, 0x42222000, 0x2222000, 0x2222000, 0x2222000, 0x2222000, 0x12222000
  dc32 0x3222, 0x0122, 0x0042, 0x0004, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222222, 0x22222221, 0x22222210, 0x44444400, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x30000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222, 0x22222, 0x22222, 0x44444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222222, 0x22222222, 0x22222221, 0x44444440, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22002222, 0x21002222, 0x10002222, 0x4444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x222220, 0x222220, 0x222220, 0x444440, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x2222200, 0x2222200, 0x2222200, 0x4444400, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22220003, 0x22220000, 0x22220000, 0x44440000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222122, 0x12222222, 0x4122222, 0x44444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22230022, 0x22100022, 0x11000022, 0x0044, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22200002, 0x22200002, 0x22200002, 0x44400004, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22200000, 0x22200000, 0x22200000, 0x44400000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0222, 0x0222, 0x0222, 0x0444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222222, 0x22222222, 0x22222222, 0x44444444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22223000, 0x22110000, 0x14000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222132, 0x11222222, 0x122222, 0x44444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22230000, 0x22200000, 0x21400000, 0x40000000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x2222223, 0x2222220, 0x1222220, 0x444440, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x32222, 0x2222, 0x4122, 0x0044, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222200, 0x22222100, 0x22211000, 0x44400000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x2222212, 0x1222222, 0x412222, 0x4444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22220022, 0x22210012, 0x21100004, 0x40000000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22212222, 0x22222222, 0x12222214, 0x4444400, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x20000000, 0x10000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22032222, 0x22002222, 0x22001422, 0x44000044, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222222, 0x22212222, 0x22102222, 0x44004444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22220, 0x22220, 0x22220, 0x44440, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222222, 0x12222222, 0x4122222, 0x44444, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22200022, 0x22100012, 0x14000001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22213222, 0x22222122, 0x22221022, 0x44440044, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22000002, 0x22300002, 0x22200002, 0x44400004, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222300, 0x22222000, 0x22222000, 0x44444000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x2222212, 0x1222222, 0x122221, 0x44440, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x22222000, 0x12222000, 0x2222000, 0x4444000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0066, 0x0077, 0x8877
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0099, 0x0077, 0x6677, 0x7777
  dc32 0x7777, 0x57797, 0x777700, 0x777700, 0x99777700, 0x77777700, 0x7777aa00, 0x77770000
  dc32 0x0000, 0x0000, 0x0005, 0x0077, 0x0077, 0x9977, 0x7777, 0xb6bb7777
  dc32 0x88a87777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77770555, 0x77770000, 0x778a0000
  dc32 0x77bb0000, 0x77000000, 0x99000000, 0x0000, 0x0000, 0x0000, 0x0000, 0xb6bb6b
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x99799900, 0x77777700, 0x7788a800
  dc32 0x0000, 0x00aa, 0x0077, 0x5577, 0x7777, 0x7777, 0x977777, 0x777777
  dc32 0x77777777, 0x77777766, 0x77777700, 0x77999900, 0x77000000, 0x77000000, 0x50000000, 0x0000
  dc32 0x77000000, 0x6b0000aa, 0x0077, 0x5577, 0x7777, 0x7777, 0x797777, 0x777777
  dc32 0x777777, 0xaa777766, 0x77777700, 0x77779900, 0x77770000, 0x77770000, 0x77050000, 0x77000000
  dc32 0x77000000, 0x66000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0xbb777777, 0x77777777, 0x77777766, 0x77777700, 0x77777758, 0x77777777, 0x77777777, 0x77777777
  dc32 0xbb777777, 0x77777777, 0x77777777, 0x77777777, 0x99777777, 0x777777, 0x777777, 0x777777
  dc32 0x88000000, 0x0000, 0x0000, 0x0000, 0x58000000, 0x77000000, 0x77000000, 0x77000000
  dc32 0x77777777, 0x88777777, 0x777777, 0x66a677, 0x0077, 0x009b, 0x0000, 0x0000
  dc32 0x0000, 0xbb000000, 0x77000000, 0x77a60000, 0x77770000, 0xb9778500, 0x777700, 0x777700
  dc32 0x777777, 0x887777, 0x7777, 0x7777, 0x7777, 0xb977, 0x0077, 0x0077
  dc32 0x77000000, 0x77bb0000, 0x77770000, 0x77776a00, 0x77777700, 0x77777700, 0x77777700, 0x77777700
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xccc5, 0xdddd
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0xeceeecee, 0xdddddddd, 0xdddddddd, 0xdddddddd
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0xee000000, 0xdd000000, 0xef000000, 0x0000
  dc32 0x7777, 0x7777, 0x8885, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0077, 0x0077, 0x0077, 0x0077, 0x00a6, 0x0000, 0x0000, 0x0000
  dc32 0x77777700, 0x77777700, 0x77778500, 0x77770000, 0xa6a60000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x5cc50000, 0xdddd0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0xecceecec, 0xdddddddd, 0xffefdddd, 0xdddd
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xccc50000, 0xdddd0000
  dc32 0xdddd, 0x00dd, 0x00dd, 0x0055, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0xdddddddd, 0xdddddddd, 0xdddddddd, 0x555555dd, 0x00dd, 0x00cc, 0x0000, 0x00cc
  dc32 0x0500, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xcccccccc, 0x0000, 0xcccccccc
  dc32 0x0000, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xcccccccc, 0x0000, 0xcccccc00
  dc32 0x0005, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xcccecccc, 0x0000, 0x0000
  dc32 0x0500, 0xdddddddd, 0xdddddddd, 0xdddd5555, 0xdddd0000, 0xeccc0000, 0x0000, 0x0000
  dc32 0x0000, 0xdddddddd, 0xdddddddd, 0x555c5555, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0005, 0xdddddddd, 0xdddddddd, 0x55555c55, 0x0000, 0x0000, 0x0000, 0xcccc
  dc32 0xdddd0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xcc000000
  dc32 0xdddd, 0x00dd, 0x00dd, 0x00dd, 0x00dd, 0xeedd, 0xdddd, 0xccdddd
  dc32 0xdddd0000, 0xdddddd00, 0xdddddd00, 0xdddddd00, 0xdddddd00, 0xdddddd00, 0xdddddd00, 0xddddee00
  dc32 0x00dd, 0x00fd, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0x555555dd, 0x00dd, 0x00cc, 0x0000
  dc32 0xdddddd00, 0xdddddd05, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xcccccccc, 0x0000
  dc32 0x0000, 0x50500505, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xcccccccc, 0x0000
  dc32 0x0000, 0x50050550, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xcccccccc, 0x0000
  dc32 0x0000, 0x5055005, 0xdddddddd, 0xdddddddd, 0xdd555555, 0xdd000000, 0xcc000000, 0x0000
  dc32 0xdddd, 0x550dffd, 0xdddd0000, 0xdddd0000, 0x55550000, 0x0000, 0x0000, 0x0000
  dc32 0xdd000000, 0xfd000000, 0x0000, 0x0000, 0x00ff, 0x00dd, 0xeedd, 0xdddd
  dc32 0xdddddd, 0x5dddddd, 0xdddddddd, 0xdddddddd, 0xdddd5555, 0xdddd0000, 0xcccc0000, 0x0000
  dc32 0xdddd0000, 0xfddd0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xeeefeeee
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xee000000
  dc32 0x0000, 0x0000, 0x555050, 0xdddddd, 0xdddddd, 0x55dddd, 0xdddd, 0xcccc
  dc32 0x0000, 0x0000, 0x55055555, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xccccccdd
  dc32 0x0000, 0x0000, 0x55505055, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd
  dc32 0x0000, 0x0000, 0x5555505, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd
  dc32 0x0000, 0x0000, 0x55055550, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xddcccccc
  dc32 0x0000, 0x0000, 0x50555500, 0xdddddd00, 0xdddddd00, 0xdd055500, 0xdd000000, 0xcc000000
  dc32 0xeeee, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0xcccc, 0xdddd, 0x55dddd, 0xdddddd, 0xdddddd, 0x500dd, 0x00dd
  dc32 0xdddddddd, 0xddddddee, 0xdddddd00, 0xdddddd55, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd
  dc32 0xdd000000, 0xee000000, 0x0000, 0x55000000, 0xdd000000, 0xdd000000, 0xdddfdffd, 0xdddddddd
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xfddfdffd, 0xdddddddd
  dc32 0x00dd, 0x00ee, 0x0000, 0x0000, 0x0000, 0x0000, 0xdfdddfdd, 0xdddddddd
  dc32 0xdddddddd, 0xeeeeeeee, 0x0000, 0x0000, 0x0000, 0x0000, 0xdddffddf, 0xdddddddd
  dc32 0xdddddddd, 0xeeeeeeee, 0x0000, 0x0000, 0x0000, 0x0000, 0xfddfdddf, 0xdddddddd
  dc32 0xdd000000, 0xee000000, 0x0000, 0x0000, 0x0000, 0x0000, 0xdfdddfdd, 0xdddddddd
  dc32 0x0000, 0x0000, 0x0000, 0x55555555, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd
  dc32 0x0000, 0x0000, 0x0000, 0x55555555, 0xdddddddd, 0xdddddddd, 0xdddd0550, 0xdddd0000
  dc32 0x0000, 0xcccccc, 0xdddddd, 0x55ddfdff, 0xdddd0000, 0xdddd0000, 0x5050000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xdf000000, 0xdd000000
  dc32 0x00cc, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0xcccccccc, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0xdddddddd, 0xdddddddd, 0xeeeeeeee, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xffffdddd, 0xdddd, 0xdddd, 0x0050
  dc32 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0x5000
  dc32 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0x500000
  dc32 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0xdddddddd, 0x50000050
  dc32 0xddcccccc, 0xdd000000, 0xddcccccc, 0xdddddddd, 0xddddddff, 0xdddddd00, 0xdddddd00, 0x0000
  dc32 0xcccccccc, 0x0000, 0xcccccccc, 0xdddddddd, 0xffffffff, 0x0000, 0x0000, 0x0000
  dc32 0xccddefef, 0xdddddd, 0xccdddddd, 0xdddddddd, 0xffffffff, 0x0000, 0x0000, 0x0000
  dc32 0xeefeeefe, 0xdddddddd, 0xddddeeee, 0xdddd0000, 0xffff0000, 0x0000, 0x0000, 0x0000
  dc32 0xdd000000, 0xdd000000, 0xee000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x99999999, 0x77777777, 0x77777777, 0x77777777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x55055777, 0x0777, 0x0777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777778, 0x777777a0
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x5557777, 0x8777, 0x0777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777705, 0x77777700, 0x777777b0
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x5507777, 0x8777, 0x0777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777778, 0x77777780
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x50555077, 0x0977, 0x6777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777775, 0x77777780, 0x77777800
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x50050550, 0x0000, 0x0000
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x50050550, 0x0000, 0x0006
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x57500550, 0x97000000, 0x77000000
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x50577775, 0x87770, 0x8770
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x5770555, 0x770000, 0x770000
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x87777777, 0x8777777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77505705, 0x77000700, 0x77000700
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x5555055, 0x0000, 0x0000
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x55050505, 0x0000, 0x0000
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777770, 0x87777780, 0x8777800
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x57055505, 0x7000000, 0xb7000000
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x55057777, 0x7777, 0x8777
  dc32 0x0777, 0x0777, 0x0777, 0x0777, 0x0777, 0x0777, 0x0777, 0x0777
  dc32 0x77777a00, 0x77777000, 0x7777b000, 0x77790000, 0x77700000, 0x77700000, 0x77500000, 0x77000000
  dc32 0x0a77, 0x0067, 0x0007, 0x5000009, 0x7000000, 0x77700000, 0x5000000, 0x0000
  dc32 0x77777776, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x7777777, 0x7777777, 0x5777777
  dc32 0x0677, 0x0077, 0x800000b7, 0x75000009, 0x77000000, 0x77000000, 0x77900000, 0x77790000
  dc32 0x77777700, 0x77777600, 0x77777000, 0x77779000, 0x77770000, 0x77700000, 0x7700000, 0x8500000
  dc32 0x7777, 0xa7777, 0x77777, 0x577777, 0x777777, 0x7777777, 0x7777777, 0x97777777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x7777777, 0x7777777, 0x5777777
  dc32 0x7777a000, 0x77770000, 0x77770000, 0x77770000, 0x77770000, 0x77770000, 0x77770000, 0x77770000
  dc32 0x0000, 0xaaa8a, 0x8877777, 0x7777777, 0x7777777, 0x77777777, 0x77777777, 0x77777777
  dc32 0x0067, 0xaaaaaa77, 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x77777500, 0x77777000
  dc32 0x77000000, 0x770000aa, 0x77000077, 0x77000079, 0x77000070, 0x77000070, 0x7000070, 0x7000079
  dc32 0x0000, 0xaaaaaaaa, 0x77777777, 0x99999999, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0770, 0x0770, 0x0770, 0x0770, 0x0770, 0x0770, 0x7770, 0x97770
  dc32 0x770000, 0x770000, 0x770000, 0x770005, 0x770007, 0x770007, 0x770007, 0x770007
  dc32 0xa77777, 0x67777, 0xb777, 0x0977, 0x0077, 0x70000007, 0x77000000, 0x77900000
  dc32 0x77000700, 0x77000700, 0x77000700, 0x77000700, 0x77000700, 0x77000700, 0x77000700, 0x57000700
  dc32 0x0000, 0xaaaaaa, 0x777777, 0x777777, 0x777777, 0x777777, 0x777777, 0x777777
  dc32 0x0000, 0xaaaa0000, 0x77778800, 0x77777700, 0x77777700, 0x77777770, 0x77777770, 0x77777770
  dc32 0xa77000, 0x77000, 0x7b000, 0x70000, 0x70000, 0x70000, 0x70000, 0x70000
  dc32 0x0000, 0x8aaa8, 0x877777, 0x5777777, 0x7777777, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0xaaaaaaaa, 0x77777777, 0x77777777, 0x77777777, 0x0000, 0x0007, 0x00b7
  dc32 0x77000000, 0x770000aa, 0x77000077, 0x77000079, 0x77000070, 0x77000070, 0x77000070, 0x77000079
  dc32 0x0777, 0x0777, 0x0777, 0x0777, 0x0777, 0x0777, 0x7777, 0x97777
  dc32 0x77777777, 0x77777777, 0x77777766, 0x77777b00, 0x77799000, 0x77700000, 0x77000000, 0x77777777
  dc32 0x77777777, 0x77777777, 0x66666666, 0x0000, 0x0000, 0x0000, 0x0000, 0x77777777
  dc32 0x0777, 0x0777, 0x0777, 0x077b, 0x5790, 0x7700, 0x77700, 0x77777777
  dc32 0x78000000, 0xa0000066, 0x0077, 0x0877, 0x5777, 0x7777, 0x7777, 0x77777777
  dc32 0x0000, 0x66666000, 0x77777a00, 0x77777780, 0x77777770, 0x77777770, 0x77777777, 0x77777777
  dc32 0x877777, 0x77777, 0x67777, 0xb777, 0x0777, 0x0777, 0x0077, 0x77777777
  dc32 0x77770000, 0x77776000, 0x77777000, 0x77777800, 0x77777750, 0x77777770, 0x77777770, 0x77777777
  dc32 0x000b, 0x0007, 0x00a7, 0x0077, 0x0877, 0x0777, 0x7777, 0x77977777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x7777777b, 0x77777790, 0x77777700, 0x77777700, 0x77777777
  dc32 0x877777, 0x77777, 0x67777, 0xb777, 0x0779, 0x0770, 0x0077, 0x77777777
  dc32 0x77777777, 0x77777777, 0x77777777, 0x77777777, 0x99999999, 0x0000, 0x0000, 0x77977777
  dc32 0x77770000, 0x77770000, 0x77770000, 0x77778000, 0x99977000, 0x97000, 0x7700, 0x77777777
  dc32 0x87777777, 0x7777777, 0x6677777, 0xbbbbb, 0x0000, 0x0000, 0x0000, 0x77777777
  dc32 0x77777000, 0x77777000, 0x77777000, 0xbbbbb000, 0x0000, 0x0000, 0x0000, 0x77777777
  dc32 0x7000077, 0x7000077, 0x7000077, 0x7000077, 0x7000077, 0x7000077, 0x7000077, 0x77777777
  dc32 0xb0000000, 0x76000000, 0x77a00000, 0x77780000, 0x77778000, 0x77777000, 0x77777700, 0x77777777
  dc32 0xbb77770, 0xa777770, 0x677770, 0xb7770, 0x9770, 0x0970, 0x0070, 0x77777777
  dc32 0x770007, 0x770007, 0x770007, 0x770007, 0x770007, 0x770007, 0x770007, 0x77777777
  dc32 0x777b0000, 0x77776000, 0x77777a00, 0x77777780, 0x77777775, 0x77777777, 0x77777777, 0x77777777
  dc32 0x8000700, 0x0700, 0x0700, 0x0700, 0x0700, 0x0700, 0x70000700, 0x77977777
  dc32 0x777777, 0x777777, 0x777777, 0xbbbbbb, 0x0000, 0x0000, 0x0000, 0x97777777
  dc32 0x77777780, 0x77777a00, 0x77776000, 0xbbbb0000, 0x0000, 0x0000, 0x0000, 0x77777777
  dc32 0x70000, 0x70000, 0x76000, 0x877000, 0x777500, 0x777700, 0x7777770, 0x77777777
  dc32 0xbbbbbbb, 0xa777777, 0x677777, 0xbbbbb, 0x0000, 0x0000, 0x0000, 0x77777777
  dc32 0xbbbbbb77, 0x77777777, 0x77777777, 0xbbbbbbbb, 0x0000, 0x0000, 0x0000, 0x77977777
  dc32 0x77000077, 0x77000077, 0x77000077, 0xb7000077, 0x7000077, 0x7000077, 0x77000077, 0x77777777
  dc32 0xb0000000, 0x76000000, 0x77a00000, 0x77780000, 0x77775000, 0x77777000, 0x77777700, 0x77777777
  dc32 0xbbb77777, 0xa7777777, 0x6777777, 0xb77777, 0x97777, 0x7777, 0x0777, 0x77777777


.align 32
  dc32 1120   ; ImageJavaGrinder_image.length
_ImageJavaGrinder_image:
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008
  dw 0x0009, 0x0000, 0x0000, 0x0000, 0x0000, 0x000a, 0x000b, 0x000c
  dw 0x0000, 0x0000, 0x0000, 0x000d, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x000e, 0x000f, 0x0010, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0011, 0x0012, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0013, 0x0014, 0x0015, 0x0016, 0x0017, 0x0018, 0x0019, 0x001a
  dw 0x001b, 0x001c, 0x001d, 0x001e, 0x001f, 0x0020, 0x0021, 0x0022
  dw 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 0x0028, 0x0029, 0x002a
  dw 0x002b, 0x002c, 0x0016, 0x002d, 0x002e, 0x002f, 0x0030, 0x0031
  dw 0x0000, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x0038
  dw 0x0039, 0x003a, 0x003b, 0x0016, 0x003c, 0x003d, 0x0019, 0x001a
  dw 0x003e, 0x003f, 0x0040, 0x0041, 0x0042, 0x0020, 0x0043, 0x0044
  dw 0x0045, 0x0046, 0x0047, 0x0048, 0x0049, 0x004a, 0x004b, 0x004c
  dw 0x004d, 0x004e, 0x0016, 0x004f, 0x0050, 0x0051, 0x0019, 0x0052
  dw 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 0x0058, 0x0059, 0x005a
  dw 0x005b, 0x005c, 0x005d, 0x005e, 0x005f, 0x0060, 0x0061, 0x0062
  dw 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 0x0068, 0x0069, 0x006a
  dw 0x006b, 0x006c, 0x006d, 0x005e, 0x006e, 0x006f, 0x0070, 0x0071
  dw 0x0072, 0x0073, 0x005e, 0x0068, 0x0074, 0x0075, 0x0061, 0x0076
  dw 0x0000, 0x0077, 0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x007e, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x007f, 0x0080, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0081, 0x0082, 0x0083, 0x0084, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0085, 0x0086, 0x0087, 0x0088, 0x0089, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x008a, 0x0000, 0x008b, 0x008c, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x008d, 0x008e, 0x008f, 0x0090, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0091
  dw 0x0092, 0x0093, 0x0000, 0x0094, 0x0095, 0x0096, 0x0000, 0x0097
  dw 0x0098, 0x0099, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x009a
  dw 0x009b, 0x009c, 0x009d, 0x009e, 0x009f, 0x00a0, 0x00a1, 0x00a2
  dw 0x00a3, 0x00a4, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x00a5, 0x00a6, 0x00a7, 0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac
  dw 0x00ad, 0x00ae, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x00af
  dw 0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x00b8, 0x00b9
  dw 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf, 0x00c0, 0x00c1
  dw 0x00c2, 0x00c3, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x00c4, 0x00c5
  dw 0x00c6, 0x00c7, 0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd
  dw 0x00ce, 0x00cf, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0
  dw 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0
  dw 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0
  dw 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0
  dw 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d2
  dw 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7, 0x00d8, 0x00d3, 0x00d4
  dw 0x00d1, 0x00d1, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00da, 0x00dd
  dw 0x00de, 0x00df, 0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e1, 0x00e2
  dw 0x00e4, 0x00e1, 0x00e5, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00e6
  dw 0x00e7, 0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00e7, 0x00e8
  dw 0x00ed, 0x00d1, 0x00ee, 0x00ef, 0x00f0, 0x00f1, 0x00f2, 0x00f3
  dw 0x00f4, 0x00f5, 0x00f6, 0x00f7, 0x00f8, 0x00f9, 0x00fa, 0x00fb
  dw 0x00fc, 0x00f2, 0x00fd, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00fe, 0x00ff, 0x0100
  dw 0x0101, 0x0102, 0x0103, 0x0104, 0x0105, 0x0106, 0x0101, 0x0102
  dw 0x0107, 0x0108, 0x0109, 0x010a, 0x010b, 0x010c, 0x010d, 0x010e
  dw 0x010f, 0x0110, 0x0111, 0x0112, 0x0113, 0x0114, 0x0115, 0x0116
  dw 0x0117, 0x0118, 0x0119, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1
  dw 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1, 0x00d1


.align 32
  dc32 16   ; ImageJavaGrinder_palette.length
_ImageJavaGrinder_palette:
  dw 0x0eee, 0x0444, 0x0000, 0x0888, 0x0222, 0x0ece, 0x066e, 0x000e
  dw 0x0aae, 0x022e, 0x088e, 0x044e, 0x0eaa, 0x0e00, 0x0e66, 0x0e22


.align 32
  dc32 936   ; ImageMike_pattern.length
_ImageMike_pattern:
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0011, 0x1110, 0x11000, 0x10000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x11000000, 0x1100000, 0x100000, 0x100000
  dc32 0x1111, 0x10001, 0x10000, 0x100000, 0x1000000, 0x1000000, 0x10000000, 0x10000000
  dc32 0x0000, 0x0000, 0x11000000, 0x100000, 0x10000, 0x1100, 0x0010, 0x0001
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0010, 0x0010, 0x0010
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x10000000, 0x10000000, 0x10000000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x11110
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x1111111, 0x10000000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0001, 0x11110, 0x11100000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x1100000, 0x11000000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0001, 0x0011, 0x0110, 0x1100, 0x1000
  dc32 0x100000, 0x1000000, 0x10000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x10000, 0x10000, 0x1000, 0x1000, 0x0100, 0x0010, 0x0010, 0x0011
  dc32 0x0001, 0x0001, 0x0010, 0x0100, 0x0100, 0x1000, 0x1000, 0x10000
  dc32 0x11000000, 0x100000, 0x10000, 0x1000, 0x1100, 0x0010, 0x0010, 0x0001
  dc32 0x0001, 0x0001, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x10000000, 0x10000000, 0x10000000, 0x10000000, 0x1000000
  dc32 0x10000000, 0x10000000, 0x10000000, 0x10000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x1110, 0x110000, 0x11000000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x11111111, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x0010, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0001, 0x0110, 0x11111000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0111, 0x11111000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0011, 0x11111100, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x11100000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x11000, 0x10000, 0x10000, 0x10000, 0x100000, 0x100000, 0x100000, 0x100000
  dc32 0x0001, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x10000, 0x10010000, 0x10100000, 0x10100000, 0x1000000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x10000000, 0x10000000, 0x1000000, 0x1100000, 0x100000, 0x100000
  dc32 0x1000000, 0x1000000, 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x10000
  dc32 0x100000, 0x100000, 0x100000, 0x100001, 0x100110, 0x11000, 0x10000, 0x11000
  dc32 0x0110, 0x111000, 0x11100000, 0x10000000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010
  dc32 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x100000
  dc32 0x100000, 0x100000, 0x10000, 0x10000, 0x10000, 0x10100, 0x11000, 0x1000
  dc32 0x10000, 0x1000, 0x1000, 0x1000, 0x1000, 0x0100, 0x0010, 0x0010
  dc32 0x11000, 0x11000, 0x1100, 0x1011, 0x0100, 0x0100, 0x0100, 0x0010
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x1100000, 0x11000, 0x0110
  dc32 0x0010, 0x0010, 0x0010, 0x0010, 0x0001, 0x0001, 0x0001, 0x0001
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0011, 0x1100
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0011, 0x111100, 0x11000000, 0x0000
  dc32 0x0000, 0x0000, 0x0111, 0x1111000, 0x10000000, 0x0000, 0x0000, 0x0000
  dc32 0x100000, 0x110000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0011, 0x0001, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x10000000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0010, 0x0010, 0x0010, 0x0011, 0x0001, 0x0000, 0x0000, 0x0000
  dc32 0x0001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x10000000, 0x1100000, 0x11100, 0x0010, 0x0001, 0x0000, 0x0000, 0x0000
  dc32 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001
  dc32 0x10000000, 0x1100000, 0x11000, 0x1110, 0x0010, 0x0000, 0x0000, 0x0000
  dc32 0x10000000, 0x10000000, 0x11000000, 0x111000, 0x0111, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x1111111, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x1111111, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x11110000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0100
  dc32 0x0000, 0x1000000, 0x100000, 0x100000, 0x100000, 0x100000, 0x10000, 0x10000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0111
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x111000, 0x11000000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x1111, 0x1110000, 0x11000000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x11111000, 0x0111, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x11100000
  dc32 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100
  dc32 0x10000, 0x10000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000
  dc32 0x0000, 0x0000, 0x0000, 0x0011, 0x11100, 0x110000, 0x1000000, 0x10000000
  dc32 0x11000, 0x1100000, 0x10000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0001, 0x0000, 0x0011, 0x0110, 0x0100
  dc32 0x0111, 0x11100, 0x1100000, 0x11000000, 0x0000, 0x10000000, 0x0000, 0x0000
  dc32 0x11000, 0x0110, 0x0001, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x1000000, 0x100000, 0x10000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0010, 0x0010, 0x0001
  dc32 0x0000, 0x10000000, 0x10000000, 0x10000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000
  dc32 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000
  dc32 0x0100, 0x0100, 0x0100, 0x0100, 0x0010, 0x0010, 0x0001, 0x0001
  dc32 0x10000000, 0x10000000, 0x10000000, 0x10000000, 0x1000000, 0x1000000, 0x110000, 0x1000
  dc32 0x0100, 0x0100, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000
  dc32 0x1000, 0x1000, 0x0100, 0x0100, 0x0010, 0x0010, 0x0001, 0x0001
  dc32 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000
  dc32 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x10000, 0x10000, 0x10000
  dc32 0x10111000, 0x10000100, 0x10000011, 0x10000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000
  dc32 0x0000, 0x0000, 0x0000, 0x10000000, 0x1000000, 0x100000, 0x11000, 0x0100
  dc32 0x10000, 0x10000, 0x10000, 0x10000, 0x10000, 0x10000, 0x10000, 0x100000
  dc32 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x1000, 0x0110, 0x0001, 0x10000000, 0x10000000, 0x10000000, 0x10000000, 0x10000000
  dc32 0x0000, 0x0000, 0x11000000, 0x110000, 0x1000, 0x1000, 0x0111, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x1000000
  dc32 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x0100
  dc32 0x10000000, 0x10000000, 0x10000000, 0x10000000, 0x10000001, 0x10000000, 0x10000000, 0x10000000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x11111110, 0x0001, 0x0000, 0x0000
  dc32 0x10000, 0x10000, 0x1000, 0x1000, 0x0100, 0x11111111, 0x0100, 0x0010
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x11100000, 0x0000, 0x0000
  dc32 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x100000, 0x10000, 0x10000
  dc32 0x0100, 0x0011, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x10000000, 0x1000000, 0x110000, 0x1000, 0x1000, 0x0110
  dc32 0x10000000, 0x10000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x110000, 0x1100, 0x0010, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x10000000, 0x1100000, 0x110000, 0x11000, 0x0000
  dc32 0x0100, 0x0010, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x10000000, 0x1000000, 0x100000, 0x10000, 0x1000
  dc32 0x1000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000, 0x1000000
  dc32 0x0010, 0x0010, 0x0001, 0x0001, 0x0001, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x10000000, 0x10000000, 0x10000000
  dc32 0x10000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1100, 0x0100, 0x0100
  dc32 0x0000, 0x10000000, 0x1100000, 0x10000, 0x10000, 0x1100, 0x0010, 0x0001
  dc32 0x1000000, 0x1000000, 0x1000000, 0x10000000, 0x10000000, 0x10000000, 0x10000000, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x10000000, 0x1100000, 0x11000, 0x0111, 0x0000
  dc32 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x11000000
  dc32 0x0010, 0x0010, 0x0010, 0x0010, 0x0100, 0x0100, 0x1000, 0x10000
  dc32 0x1000000, 0x1000000, 0x1100000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x1000000, 0x1000000, 0x11000000, 0x1100000, 0x100000, 0x100000, 0x0000, 0x0000
  dc32 0x0100, 0x0100, 0x0100, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x10000001, 0x1000001, 0x100010, 0x11010, 0x1110, 0x0000, 0x0000, 0x0000
  dc32 0x111000, 0x0110, 0x0001, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x0000, 0x0000, 0x11111111, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dc32 0x10000, 0x1100000, 0x11000000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000


.align 32
  dc32 1120   ; ImageMike_image.length
_ImageMike_image:
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x0002, 0x0000
  dw 0x0003, 0x0004, 0x0000, 0x0000, 0x0005, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0006, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0007, 0x0008, 0x0009, 0x000a, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x000b, 0x000c, 0x000d, 0x000e
  dw 0x0000, 0x0000, 0x000f, 0x0000, 0x0010, 0x0011, 0x0000, 0x0000
  dw 0x0000, 0x0012, 0x0013, 0x0014, 0x0015, 0x0000, 0x0000, 0x0016
  dw 0x0017, 0x0018, 0x0019, 0x001a, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x001b, 0x0000, 0x001c, 0x001d
  dw 0x0000, 0x0000, 0x001c, 0x001e, 0x0000, 0x001f, 0x0000, 0x0000
  dw 0x0000, 0x0020, 0x0021, 0x0000, 0x0000, 0x0000, 0x0000, 0x0022
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0023, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0024, 0x0000, 0x0025, 0x0000, 0x0000
  dw 0x0000, 0x0026, 0x0027, 0x0000, 0x0000, 0x0000, 0x0000, 0x0028
  dw 0x0029, 0x002a, 0x002b, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x002c, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x002d, 0x002e, 0x0000
  dw 0x0000, 0x002f, 0x0030, 0x0031, 0x0027, 0x0000, 0x0000, 0x0032
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0030, 0x0033, 0x0000, 0x0000
  dw 0x0034, 0x0035, 0x0036, 0x0037, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0038, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0039, 0x0000
  dw 0x003a, 0x003b, 0x0000, 0x0000, 0x0000, 0x0013, 0x003c, 0x003d
  dw 0x003e, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x003f, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0040, 0x0041
  dw 0x0042, 0x0000, 0x0000, 0x0000, 0x0043, 0x0044, 0x0000, 0x0000
  dw 0x0045, 0x0046, 0x0000, 0x0000, 0x0000, 0x0047, 0x0000, 0x0000
  dw 0x001c, 0x0048, 0x0000, 0x0000, 0x003e, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0049, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x004a, 0x004b
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x004c, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x004d, 0x0000, 0x0000, 0x0000, 0x004e, 0x004f, 0x0000
  dw 0x0000, 0x0050, 0x0000, 0x0000, 0x0051, 0x0052, 0x0000, 0x0000
  dw 0x0000, 0x0053, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0054, 0x0055
  dw 0x0056, 0x0057, 0x0000, 0x0000, 0x0058, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0032, 0x0000, 0x0000, 0x0000, 0x0000, 0x0059, 0x0015
  dw 0x005a, 0x005b, 0x005c, 0x0000, 0x005d, 0x005e, 0x005f, 0x0000
  dw 0x0000, 0x0023, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0060
  dw 0x0000, 0x0061, 0x0062, 0x0000, 0x0063, 0x0064, 0x0000, 0x0000
  dw 0x0000, 0x0032, 0x0000, 0x0000, 0x0000, 0x0000, 0x0065, 0x0000
  dw 0x0000, 0x0066, 0x0067, 0x0000, 0x0068, 0x0000, 0x0030, 0x0069
  dw 0x0013, 0x006a, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0063, 0x006b, 0x006c
  dw 0x0000, 0x006d, 0x0000, 0x0000, 0x0000, 0x0000, 0x006e, 0x0000
  dw 0x0000, 0x0000, 0x006f, 0x0000, 0x0070, 0x0000, 0x0000, 0x0000
  dw 0x0071, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0072
  dw 0x0073, 0x0074, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000


.align 32
  dc32 16   ; ImageMike_palette.length
_ImageMike_palette:
  dw 0x0000, 0x0808, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
  dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000


.align 32
  dc32 16   ; Mandelbrots_palette.length
_Mandelbrots_palette:
  dw 0x0000, 0x004e, 0x0044, 0x0040, 0x0440, 0x0880, 0x0808, 0x0e0e
  dw 0x0e4e, 0x002e, 0x088e, 0x0444, 0x0e40, 0x0ee0, 0x0400, 0x0000


.align 32
  dc32 198   ; PlaySong_z80_code.length
_PlaySong_z80_code:
  db 0x31, 0x00, 0x20, 0xdd, 0x21, 0xaa, 0x00, 0x0e
  db 0x30, 0x06, 0x1c, 0xcd, 0x94, 0x00, 0xcd, 0x94
  db 0x00, 0xcd, 0x94, 0x00, 0x0c, 0xdd, 0x23, 0x10
  db 0xf2, 0x3e, 0xb4, 0x32, 0x00, 0x40, 0x3e, 0xc0
  db 0x32, 0x01, 0x40, 0x3e, 0xa4, 0x32, 0x00, 0x40
  db 0x3e, 0x22, 0x32, 0x01, 0x40, 0x3e, 0xa0, 0x32
  db 0x00, 0x40, 0x3e, 0x69, 0x32, 0x01, 0x40, 0x3e
  db 0xb5, 0x32, 0x00, 0x40, 0x3e, 0xc0, 0x32, 0x01
  db 0x40, 0x3e, 0xa5, 0x32, 0x00, 0x40, 0x3e, 0x23
  db 0x32, 0x01, 0x40, 0x3e, 0xa1, 0x32, 0x00, 0x40
  db 0x3e, 0x09, 0x32, 0x01, 0x40, 0x3e, 0xb4, 0x32
  db 0x02, 0x40, 0x3e, 0xc0, 0x32, 0x03, 0x40, 0x3e
  db 0xa4, 0x32, 0x02, 0x40, 0x3e, 0x23, 0x32, 0x03
  db 0x40, 0x3e, 0xa0, 0x32, 0x02, 0x40, 0x3e, 0x9c
  db 0x32, 0x03, 0x40, 0x3e, 0x28, 0x32, 0x00, 0x40
  db 0x3e, 0xf0, 0x32, 0x01, 0x40, 0x3e, 0x28, 0x32
  db 0x00, 0x40, 0x3e, 0xf1, 0x32, 0x01, 0x40, 0x3e
  db 0x28, 0x32, 0x00, 0x40, 0x3e, 0xf4, 0x32, 0x01
  db 0x40, 0xc3, 0x91, 0x00, 0x79, 0x32, 0x00, 0x40
  db 0xdd, 0x7e, 0x00, 0x32, 0x01, 0x40, 0x79, 0x32
  db 0x02, 0x40, 0xdd, 0x7e, 0x00, 0x32, 0x03, 0x40
  db 0x0c, 0xc9, 0x71, 0x0d, 0x33, 0x01, 0x23, 0x2d
  db 0x26, 0x00, 0x5f, 0x99, 0x5f, 0x94, 0x05, 0x05
  db 0x05, 0x07, 0x02, 0x02, 0x02, 0x02, 0x11, 0x11
  db 0x11, 0xa6, 0x00, 0x00, 0x00, 0x00

.align 32
  dc32 6058   ; PlayTitleSample_z80_code.length
_PlayTitleSample_z80_code:
  db 0x3e, 0x2b, 0x32, 0x00, 0x40, 0x3e, 0x80, 0x32
  db 0x01, 0x40, 0xdd, 0x21, 0x3a, 0x00, 0x21, 0x70
  db 0x17, 0x3e, 0x2a, 0x32, 0x00, 0x40, 0xdd, 0x7e
  db 0x00, 0x32, 0x01, 0x40, 0xdd, 0x23, 0x06, 0x3e
  db 0x10, 0xfe, 0x2b, 0x7d, 0xfe, 0x00, 0x20, 0xe9
  db 0x7c, 0xfe, 0x00, 0x20, 0xe4, 0x3e, 0x2b, 0x32
  db 0x00, 0x40, 0x3e, 0x00, 0x32, 0x01, 0x40, 0xc3
  db 0x37, 0x00, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x7f, 0x7f, 0x7f, 0x7f, 0x80
  db 0x80, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x7f, 0x81
  db 0x81, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x81, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x81, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x80, 0x81
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x7f, 0x80, 0x80, 0x81, 0x80, 0x80, 0x80, 0x80
  db 0x7f, 0x81, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x81, 0x80, 0x80, 0x7f, 0x80
  db 0x80, 0x80, 0x81, 0x80, 0x80, 0x7f, 0x80, 0x80
  db 0x80, 0x7f, 0x7f, 0x80, 0x7f, 0x80, 0x80, 0x80
  db 0x7f, 0x80, 0x81, 0x80, 0x80, 0x80, 0x80, 0x81
  db 0x7f, 0x81, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x81, 0x80, 0x81, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x7f, 0x80, 0x81, 0x80, 0x7f, 0x80, 0x7f, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x7f, 0x80
  db 0x80, 0x81, 0x81, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x81, 0x80, 0x80, 0x80, 0x81, 0x80, 0x80, 0x80
  db 0x80, 0x7f, 0x80, 0x7f, 0x7f, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x81
  db 0x81, 0x7f, 0x80, 0x80, 0x80, 0x81, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x80
  db 0x7f, 0x80, 0x80, 0x80, 0x81, 0x80, 0x80, 0x80
  db 0x7f, 0x80, 0x81, 0x81, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x81
  db 0x80, 0x80, 0x7f, 0x80, 0x7f, 0x80, 0x80, 0x80
  db 0x80, 0x7f, 0x7f, 0x80, 0x80, 0x80, 0x7f, 0x81
  db 0x81, 0x80, 0x7f, 0x7f, 0x80, 0x80, 0x80, 0x80
  db 0x7f, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x81, 0x7f, 0x80, 0x81, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x81, 0x80, 0x81, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x81
  db 0x80, 0x81, 0x81, 0x80, 0x7f, 0x80, 0x80, 0x80
  db 0x80, 0x7f, 0x80, 0x81, 0x80, 0x80, 0x81, 0x7f
  db 0x80, 0x80, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x81, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x7f, 0x81, 0x80, 0x80, 0x80
  db 0x81, 0x7f, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x7f
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x81, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x81, 0x80, 0x80, 0x81, 0x80, 0x80, 0x7f, 0x81
  db 0x80, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x7f, 0x80, 0x7f, 0x7f, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x7f, 0x80, 0x80, 0x81
  db 0x80, 0x81, 0x80, 0x80, 0x80, 0x81, 0x80, 0x7f
  db 0x80, 0x80, 0x81, 0x81, 0x80, 0x81, 0x80, 0x80
  db 0x80, 0x80, 0x80, 0x81, 0x80, 0x80, 0x80, 0x81
  db 0x80, 0x80, 0x81, 0x80, 0x81, 0x80, 0x81, 0x80
  db 0x80, 0x80, 0x80, 0x80, 0x80, 0x81, 0x81, 0x80
  db 0x80, 0x80, 0x81, 0x80, 0x80, 0x81, 0x80, 0x80
  db 0x81, 0x80, 0x80, 0x80, 0x7f, 0x7f, 0x80, 0x80
  db 0x81, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x81, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x7f, 0x7f, 0x81, 0x80, 0x81, 0x7f
  db 0x81, 0x80, 0x7f, 0x80, 0x80, 0x80, 0x80, 0x80
  db 0x80, 0x80, 0x82, 0x81, 0x7f, 0x7d, 0x7b, 0x80
  db 0x7d, 0x81, 0x7b, 0x7e, 0x7d, 0x8a, 0x75, 0x5f
  db 0x6f, 0x7b, 0x98, 0xa1, 0xae, 0xa8, 0x9f, 0x93
  db 0x8a, 0x80, 0x7b, 0x73, 0x6b, 0x6c, 0x71, 0x75
  db 0x74, 0x79, 0x79, 0x79, 0x79, 0x77, 0x79, 0x7a
  db 0x7a, 0x78, 0x7a, 0x7d, 0x83, 0x85, 0x88, 0x8a
  db 0x88, 0x87, 0x85, 0x82, 0x81, 0x81, 0x81, 0x83
  db 0x85, 0x88, 0x88, 0x89, 0x88, 0x87, 0x86, 0x84
  db 0x83, 0x83, 0x82, 0x81, 0x80, 0x7e, 0x7d, 0x7c
  db 0x7d, 0x7d, 0x7e, 0x7e, 0x80, 0x80, 0x7f, 0x80
  db 0x7f, 0x81, 0x7d, 0x80, 0x7f, 0x7f, 0x81, 0x81
  db 0x82, 0x81, 0x7d, 0x82, 0x7f, 0x7d, 0x81, 0x7f
  db 0x7e, 0x81, 0x80, 0x81, 0x7f, 0x7f, 0x82, 0x7b
  db 0x80, 0x7f, 0x81, 0x7f, 0x7e, 0x80, 0x7c, 0x7e
  db 0x80, 0x82, 0x80, 0x85, 0x7e, 0x81, 0x7e, 0x7f
  db 0x82, 0x82, 0x82, 0x80, 0x85, 0x7e, 0x82, 0x7f
  db 0x82, 0x84, 0x7e, 0x83, 0x81, 0x80, 0x7f, 0x83
  db 0x7d, 0x7d, 0x7e, 0x82, 0x7d, 0x7a, 0x82, 0x7b
  db 0x7a, 0x7f, 0x7c, 0x85, 0x82, 0x81, 0x85, 0x81
  db 0x86, 0x83, 0x80, 0x84, 0x81, 0x80, 0x7b, 0x7c
  db 0x7d, 0x80, 0x86, 0x7f, 0x79, 0x74, 0x80, 0x84
  db 0x8a, 0x85, 0x87, 0x8a, 0x82, 0x8a, 0x85, 0x80
  db 0x7d, 0x7a, 0x76, 0x78, 0x76, 0x79, 0x7d, 0x7d
  db 0x7f, 0x7a, 0x80, 0x7d, 0x81, 0x80, 0x83, 0x8a
  db 0x81, 0x7f, 0x72, 0x74, 0x6e, 0x74, 0x81, 0x87
  db 0x8d, 0x86, 0x91, 0x79, 0x73, 0x7d, 0x87, 0x97
  db 0x97, 0x9b, 0x8e, 0x8a, 0x87, 0x81, 0x83, 0x83
  db 0x7d, 0x7e, 0x72, 0x77, 0x69, 0x64, 0x6c, 0x65
  db 0x7a, 0x8a, 0x86, 0x8a, 0x87, 0x8d, 0x83, 0x77
  db 0x74, 0x63, 0x59, 0x55, 0x5c, 0x5d, 0x64, 0x6e
  db 0x95, 0x9e, 0xb2, 0xb8, 0xc8, 0xb4, 0xb3, 0xc5
  db 0x98, 0x9c, 0x88, 0x72, 0x6a, 0x5e, 0x43, 0x42
  db 0x32, 0x3c, 0x58, 0x5e, 0x6c, 0x8e, 0x9e, 0xad
  db 0xcc, 0xd6, 0xf4, 0xed, 0xf7, 0xc9, 0xa2, 0x6d
  db 0x16, 0x13, 0x0f, 0x0d, 0x18, 0x2a, 0x65, 0x6e
  db 0x55, 0x66, 0x5f, 0x75, 0xb2, 0xea, 0xf5, 0xec
  db 0xd6, 0xbb, 0x58, 0x0b, 0x31, 0x00, 0x33, 0x93
  db 0x95, 0xb9, 0x8d, 0x44, 0x59, 0x76, 0x9d, 0xe6
  db 0xf7, 0xed, 0xe4, 0xb8, 0x2e, 0x0f, 0x32, 0x19
  db 0x8e, 0xf5, 0xc0, 0xba, 0x5e, 0x03, 0x1e, 0x1c
  db 0x90, 0xf6, 0xdd, 0xb2, 0xa7, 0x4c, 0x04, 0x21
  db 0x38, 0x51, 0xc9, 0xf6, 0xce, 0x6f, 0x15, 0x48
  db 0x4c, 0x96, 0xee, 0xe2, 0xde, 0xa6, 0x6f, 0x39
  db 0x1b, 0x48, 0xac, 0xdb, 0xb3, 0xb5, 0x73, 0x0e
  db 0x2b, 0x46, 0x63, 0xde, 0xdb, 0xc0, 0xbe, 0x5b
  db 0x18, 0x44, 0x39, 0x74, 0xe1, 0xe1, 0xaf, 0x55
  db 0x4f, 0x44, 0x56, 0x93, 0xbb, 0xd0, 0xca, 0xa3
  db 0xb8, 0x43, 0x15, 0x4d, 0x5b, 0xb9, 0xb8, 0xa3
  db 0x43, 0x40, 0x62, 0x55, 0xbc, 0xd5, 0x8e, 0x9b
  db 0x7d, 0x50, 0x98, 0xd5, 0x55, 0x48, 0xa3, 0x3a
  db 0x51, 0xc9, 0xcb, 0x42, 0x6b, 0x93, 0x3e, 0x6a
  db 0xa2, 0xa2, 0xb0, 0xb7, 0x9a, 0x99, 0x77, 0x1d
  db 0x1a, 0x75, 0x76, 0x7e, 0xcb, 0x91, 0x36, 0x72
  db 0x79, 0x4d, 0x8d, 0xe7, 0x7f, 0xa5, 0xc8, 0x39
  db 0x72, 0xb7, 0x60, 0x2c, 0x9a, 0x7f, 0x47, 0xab
  db 0xb3, 0x54, 0x50, 0x92, 0x7c, 0x5f, 0xb1, 0x9a
  db 0x4e, 0x75, 0x9f, 0xa2, 0xc8, 0x8f, 0x4b, 0x47
  db 0x9d, 0x8a, 0x8d, 0xc8, 0x6e, 0x43, 0x7f, 0xa5
  db 0x3d, 0x89, 0xc1, 0x7e, 0x8b, 0xb7, 0x84, 0x92
  db 0x5b, 0x23, 0x50, 0x95, 0x9b, 0x82, 0x87, 0x39
  db 0x4a, 0x87, 0xbe, 0x89, 0x78, 0x5e, 0x60, 0x88
  db 0xc1, 0xa5, 0xa3, 0xa6, 0x4a, 0x4a, 0x79, 0xbf
  db 0x68, 0x8c, 0x8b, 0x3c, 0x56, 0xa1, 0xb4, 0x57
  db 0x6b, 0xad, 0x96, 0x98, 0xaa, 0x9a, 0x58, 0x25
  db 0x8c, 0xbf, 0x91, 0x92, 0x74, 0x32, 0x4d, 0xa3
  db 0xbb, 0x6d, 0x51, 0x56, 0x65, 0xa8, 0xd1, 0x9f
  db 0xa1, 0x80, 0x31, 0x56, 0xa8, 0xbd, 0x75, 0x59
  db 0x61, 0x57, 0x8c, 0xc8, 0x7c, 0x4a, 0x8a, 0xbc
  db 0x8d, 0xa0, 0xa7, 0x52, 0x38, 0x7b, 0xb7, 0x78
  db 0x82, 0x80, 0x35, 0x5b, 0xab, 0xb2, 0x58, 0x4f
  db 0x6e, 0x70, 0xb5, 0xc8, 0xa7, 0x73, 0x48, 0x57
  db 0xa8, 0xbf, 0x9a, 0x5e, 0x4a, 0x5c, 0x8d, 0xc8
  db 0x97, 0x4c, 0x5c, 0xa5, 0xa2, 0x93, 0xa0, 0x70
  db 0x28, 0x64, 0xb1, 0x9c, 0x6e, 0x73, 0x56, 0x46
  db 0x95, 0xcb, 0x83, 0x45, 0x76, 0x71, 0x84, 0xc2
  db 0xbc, 0x77, 0x4a, 0x67, 0xaf, 0xb4, 0xa3, 0x6f
  db 0x3b, 0x55, 0x8e, 0xbb, 0x96, 0x55, 0x48, 0x6c
  db 0xa1, 0xb5, 0x9a, 0xa2, 0x60, 0x2f, 0x7f, 0xab
  db 0x90, 0x53, 0x68, 0x68, 0x54, 0xb2, 0xbb, 0x63
  db 0x56, 0xb1, 0xa1, 0x5e, 0xa6, 0xad, 0x56, 0x55
  db 0x93, 0x81, 0x7b, 0xb9, 0x90, 0x44, 0x5e, 0xab
  db 0x9a, 0x7b, 0x86, 0x60, 0x62, 0x8b, 0xbd, 0x88
  db 0x89, 0x9e, 0x4d, 0x49, 0x8a, 0xbc, 0x71, 0x58
  db 0x7b, 0x56, 0x63, 0xba, 0x9d, 0x4a, 0x80, 0xba
  db 0x8a, 0x71, 0xa2, 0x75, 0x3e, 0x76, 0xb1, 0x88
  db 0x69, 0x83, 0x61, 0x57, 0x9e, 0xba, 0x71, 0x59
  db 0x82, 0x78, 0x80, 0xae, 0x9b, 0x5b, 0x9c, 0xad
  db 0x52, 0x50, 0xa2, 0xad, 0x5e, 0x6d, 0x92, 0x60
  db 0x60, 0xaa, 0x7a, 0x50, 0x95, 0xbc, 0x98, 0x9e
  db 0x7c, 0x36, 0x60, 0xa9, 0xa9, 0x6c, 0x61, 0x54
  db 0x57, 0x9a, 0xc5, 0x86, 0x4f, 0x6b, 0x8d, 0xa7
  db 0xb1, 0xa0, 0x60, 0x3e, 0x77, 0xb0, 0xa7, 0x98
  db 0x5a, 0x40, 0x5f, 0xa3, 0xb0, 0x61, 0x52, 0x66
  db 0x88, 0xb6, 0xbb, 0x74, 0x4e, 0x72, 0xb5, 0xa3
  db 0x93, 0x86, 0x47, 0x46, 0x83, 0xb3, 0x8c, 0x92
  db 0x84, 0x3f, 0x5a, 0xa5, 0x98, 0x4c, 0x5f, 0x95
  db 0x74, 0x8a, 0xb3, 0x72, 0x50, 0x8b, 0xaa, 0x78
  db 0x6e, 0x8e, 0x70, 0x7b, 0xae, 0xb4, 0x88, 0x50
  db 0x57, 0x97, 0xb9, 0x9d, 0x68, 0x4d, 0x56, 0x87
  db 0xa8, 0x77, 0x76, 0x91, 0x8d, 0xa7, 0x9b, 0x49
  db 0x4c, 0x99, 0xb6, 0x7e, 0x72, 0x77, 0x45, 0x61
  db 0xb1, 0x9a, 0x72, 0xa2, 0x70, 0x4e, 0x9d, 0xa5
  db 0x65, 0x45, 0x96, 0xa4, 0x70, 0xa5, 0x84, 0x45
  db 0x6f, 0xaf, 0x95, 0x4e, 0x77, 0x78, 0x6c, 0xa9
  db 0xc1, 0x92, 0x4a, 0x50, 0x99, 0xb3, 0x99, 0x7f
  db 0x55, 0x44, 0x73, 0xaf, 0x7b, 0x61, 0x99, 0x99
  db 0xa3, 0x96, 0x46, 0x46, 0x8f, 0xb1, 0x8e, 0x72
  db 0x6f, 0x4f, 0x66, 0xb8, 0xa0, 0x5c, 0x7d, 0x72
  db 0x62, 0xae, 0xb6, 0x6d, 0x44, 0x71, 0xbc, 0xa8
  db 0xa1, 0x83, 0x40, 0x58, 0x9e, 0xb3, 0x64, 0x49
  db 0x5e, 0x6b, 0xa4, 0xc7, 0x9d, 0x53, 0x4b, 0x95
  db 0xbe, 0x9b, 0x97, 0x69, 0x3b, 0x5d, 0xae, 0x79
  db 0x46, 0xad, 0x9d, 0x86, 0xa0, 0x64, 0x2e, 0x74
  db 0xb5, 0x81, 0x5e, 0x77, 0x61, 0x6c, 0xb1, 0xb2
  db 0x61, 0x5f, 0x86, 0x92, 0xa9, 0xb5, 0x9b, 0x63
  db 0x42, 0x67, 0xac, 0xaf, 0x85, 0x4a, 0x58, 0x6a
  db 0xa9, 0xb1, 0x6b, 0x50, 0x69, 0xa3, 0xc5, 0x89
  db 0x4e, 0x58, 0x9b, 0xb4, 0x91, 0x9a, 0x57, 0x3f
  db 0x6c, 0x99, 0x86, 0x8d, 0xaa, 0x72, 0x71, 0x8c
  db 0x51, 0x49, 0x99, 0xa8, 0x5b, 0x57, 0x84, 0x6b
  db 0x84, 0xb8, 0x90, 0x57, 0x70, 0x9d, 0x9c, 0x91
  db 0xaa, 0xa3, 0x68, 0x3f, 0x69, 0xab, 0xac, 0x71
  db 0x58, 0x59, 0x5c, 0xa5, 0xbc, 0x6f, 0x53, 0x72
  db 0x9d, 0xbb, 0x9f, 0x61, 0x4a, 0x81, 0xbc, 0xa1
  db 0x9b, 0x6d, 0x40, 0x50, 0x8e, 0x9b, 0x6d, 0x9e
  db 0x9f, 0x95, 0x82, 0x46, 0x3f, 0x8a, 0xa7, 0x77
  db 0x63, 0x67, 0x5f, 0x7b, 0xbc, 0xa0, 0x5e, 0x5e
  db 0x86, 0x9a, 0xa4, 0xae, 0xa5, 0x67, 0x41, 0x67
  db 0xa4, 0xb0, 0x74, 0x4c, 0x5e, 0x6e, 0xa5, 0xbe
  db 0x7a, 0x52, 0x64, 0x9d, 0xb6, 0xb5, 0x72, 0x46
  db 0x5f, 0xa7, 0xb2, 0x9f, 0x76, 0x48, 0x4d, 0x7f
  db 0xb7, 0x5d, 0x53, 0xae, 0xb0, 0x9e, 0x8d, 0x46
  db 0x41, 0x84, 0xae, 0x9c, 0x65, 0x4d, 0x5d, 0x7b
  db 0xbd, 0xa7, 0x5c, 0x5b, 0x74, 0xa0, 0xb5, 0xaf
  db 0x93, 0x59, 0x46, 0x74, 0xa4, 0xa6, 0x6b, 0x50
  db 0x5d, 0x72, 0xa2, 0xb9, 0x6f, 0x5b, 0x6b, 0xa1
  db 0xb3, 0xa4, 0x89, 0x4e, 0x5b, 0x8e, 0xaa, 0x8d
  db 0x77, 0x4e, 0x55, 0x6f, 0xba, 0xa1, 0x4f, 0x6a
  db 0xae, 0xae, 0x9b, 0x8f, 0x48, 0x4b, 0x7c, 0xaf
  db 0x9a, 0x55, 0x4e, 0x67, 0x7c, 0xc4, 0xaf, 0x69
  db 0x5b, 0x6f, 0xab, 0xb6, 0xa1, 0x83, 0x46, 0x4a
  db 0x84, 0xb2, 0x9c, 0x4e, 0x52, 0x5d, 0x88, 0xbe
  db 0xac, 0x65, 0x50, 0x6c, 0xa6, 0xb3, 0x9b, 0x8e
  db 0x4b, 0x47, 0x89, 0xb6, 0x8b, 0x5a, 0x59, 0x60
  db 0x71, 0xb9, 0xaf, 0x6c, 0x52, 0x8e, 0xbc, 0x9d
  db 0xa2, 0x73, 0x3b, 0x5c, 0xa4, 0xab, 0x6b, 0x49
  db 0x62, 0x65, 0x99, 0xcb, 0x92, 0x57, 0x5b, 0x9b
  db 0xb5, 0xa2, 0x9c, 0x69, 0x3e, 0x57, 0x93, 0xaf
  db 0x79, 0x4b, 0x5c, 0x6b, 0x98, 0xbb, 0xa2, 0x64
  db 0x53, 0x7d, 0xac, 0xb3, 0x93, 0x9b, 0x71, 0x3a
  db 0x64, 0xa2, 0xa0, 0x54, 0x4e, 0x74, 0x6b, 0x95
  db 0xbd, 0x8f, 0x5b, 0x6c, 0xa5, 0xb2, 0x99, 0x9d
  db 0x6e, 0x36, 0x5c, 0x9c, 0xa6, 0x61, 0x4b, 0x61
  db 0x67, 0x98, 0xc5, 0x9c, 0x5f, 0x69, 0xa6, 0xb3
  db 0x98, 0x9a, 0x74, 0x40, 0x4c, 0x8c, 0xaa, 0x8c
  db 0x6d, 0x53, 0x57, 0x84, 0xb9, 0xa4, 0x64, 0x59
  db 0x64, 0x8b, 0xb5, 0xa8, 0x66, 0x88, 0x98, 0x4e
  db 0x66, 0xa4, 0xa4, 0x5e, 0x60, 0x90, 0x73, 0x72
  db 0xb3, 0x90, 0x55, 0x5a, 0x97, 0xb6, 0x8f, 0x9a
  db 0x83, 0x3b, 0x54, 0x9f, 0xae, 0x6e, 0x49, 0x68
  db 0x66, 0x88, 0xc7, 0xa0, 0x5b, 0x5e, 0x98, 0xb2
  db 0x91, 0xa1, 0x79, 0x4c, 0x53, 0x7c, 0xa0, 0xa1
  db 0x81, 0x53, 0x53, 0x7b, 0xac, 0xa5, 0x78, 0x5d
  db 0x5b, 0x75, 0xb2, 0xb2, 0x6f, 0x73, 0xb2, 0x6e
  db 0x4d, 0x8b, 0xb4, 0x77, 0x52, 0x92, 0x7c, 0x57
  db 0x98, 0xb8, 0x6a, 0x4c, 0x81, 0xb7, 0xa0, 0x95
  db 0x95, 0x54, 0x36, 0x79, 0xad, 0x8a, 0x4f, 0x64
  db 0x67, 0x62, 0xaf, 0xbd, 0x7d, 0x56, 0x81, 0xaa
  db 0x8a, 0xa9, 0xa1, 0x52, 0x4b, 0x6a, 0x8c, 0x9a
  db 0xa4, 0x76, 0x4d, 0x64, 0x9b, 0xab, 0x88, 0x6c
  db 0x63, 0x5c, 0x87, 0xc2, 0x98, 0x61, 0x97, 0x8e
  db 0x52, 0x69, 0xac, 0x9f, 0x4e, 0x79, 0x8d, 0x54
  db 0x71, 0xb8, 0x93, 0x56, 0x65, 0xa6, 0xb4, 0x93
  db 0x9a, 0x73, 0x38, 0x60, 0xa3, 0x9f, 0x55, 0x51
  db 0x6f, 0x58, 0x8e, 0xc0, 0x9e, 0x5f, 0x67, 0xa2
  db 0x95, 0x8f, 0xb2, 0x81, 0x4c, 0x54, 0x84, 0xad
  db 0xa5, 0x81, 0x51, 0x4d, 0x80, 0xaf, 0xa5, 0x77
  db 0x59, 0x58, 0x74, 0xb0, 0xa7, 0x6f, 0x98, 0x8d
  db 0x46, 0x65, 0xa6, 0xaa, 0x55, 0x66, 0x8c, 0x5c
  db 0x7f, 0xbb, 0x95, 0x53, 0x63, 0xa7, 0xb3, 0x99
  db 0x9a, 0x76, 0x38, 0x59, 0x9e, 0xa2, 0x63, 0x50
  db 0x63, 0x5f, 0x94, 0xc1, 0x98, 0x58, 0x62, 0x8f
  db 0x9c, 0xa7, 0xb1, 0x70, 0x4b, 0x62, 0x7d, 0xa0
  db 0xb3, 0x8b, 0x4e, 0x52, 0x85, 0xa6, 0xa4, 0x8a
  db 0x60, 0x52, 0x76, 0xb0, 0xa4, 0x72, 0x90, 0x9a
  db 0x48, 0x4e, 0x99, 0xab, 0x60, 0x6d, 0x8c, 0x56
  db 0x69, 0xad, 0xa3, 0x5e, 0x59, 0x93, 0xba, 0x9a
  db 0x92, 0x88, 0x4f, 0x45, 0x86, 0xa9, 0x86, 0x65
  db 0x66, 0x5c, 0x7c, 0xb6, 0xb3, 0x6a, 0x5c, 0x63
  db 0x88, 0xb1, 0xaa, 0x8c, 0x68, 0x71, 0x59, 0x7f
  db 0xba, 0xa3, 0x5c, 0x5d, 0x7a, 0x78, 0xa5, 0xb4
  db 0x7c, 0x45, 0x5f, 0x9b, 0x9e, 0x96, 0xa5, 0x85
  db 0x43, 0x51, 0x9c, 0xad, 0x6e, 0x7d, 0x73, 0x4d
  db 0x71, 0xb5, 0x9f, 0x49, 0x57, 0x87, 0xa6, 0x9c
  db 0xac, 0x7f, 0x41, 0x56, 0xa5, 0xa8, 0x99, 0x83
  db 0x48, 0x43, 0x82, 0xb2, 0xa2, 0x5d, 0x56, 0x60
  db 0x72, 0xb9, 0xb9, 0xa0, 0x5d, 0x52, 0x6b, 0xa2
  db 0xbe, 0x94, 0x58, 0x57, 0x61, 0x91, 0xc0, 0xa1
  db 0x5f, 0x53, 0x67, 0x90, 0xb7, 0xab, 0x82, 0x47
  db 0x59, 0x97, 0xad, 0x96, 0x93, 0x4c, 0x3d, 0x68
  db 0xad, 0x8d, 0x4b, 0x6e, 0x7b, 0x92, 0xb8, 0xa3
  db 0x64, 0x4f, 0x7f, 0xae, 0x97, 0xa0, 0x81, 0x49
  db 0x4f, 0x91, 0xa6, 0x76, 0x88, 0x77, 0x4a, 0x7c
  db 0xac, 0xa0, 0x62, 0x5a, 0x65, 0x7f, 0xad, 0xbb
  db 0x94, 0x55, 0x5b, 0x78, 0x9c, 0xb0, 0x9a, 0x6a
  db 0x53, 0x76, 0xab, 0xaf, 0x67, 0x4d, 0x68, 0x7c
  db 0x91, 0xb5, 0xab, 0x63, 0x52, 0x7a, 0x8b, 0x67
  db 0x9c, 0xb1, 0x79, 0x84, 0xa4, 0x7f, 0x3c, 0x65
  db 0x9f, 0x73, 0x70, 0xa7, 0x76, 0x41, 0x6e, 0xa6
  db 0x62, 0x5d, 0xb8, 0x8a, 0x6e, 0xa3, 0x9e, 0x60
  db 0x43, 0x87, 0x9b, 0x73, 0xa3, 0x9c, 0x56, 0x4e
  db 0x86, 0x8b, 0x7a, 0xa1, 0x97, 0x5a, 0x71, 0xa6
  db 0xb0, 0x80, 0x4b, 0x72, 0x72, 0x72, 0xab, 0xb6
  db 0x7c, 0x50, 0x6e, 0x7c, 0x66, 0x9b, 0xc2, 0x93
  db 0x6f, 0xa0, 0x9b, 0x52, 0x4f, 0x95, 0x8a, 0x56
  db 0x95, 0xa1, 0x5f, 0x52, 0x8e, 0x88, 0x54, 0x83
  db 0xae, 0x87, 0x8f, 0xa4, 0x8b, 0x4c, 0x4c, 0x8c
  db 0x78, 0x7e, 0xad, 0x9c, 0x5c, 0x50, 0x80, 0x72
  db 0x7b, 0xb2, 0xb0, 0x70, 0x69, 0xad, 0x92, 0x50
  db 0x65, 0x9f, 0x7d, 0x60, 0x9a, 0xa3, 0x62, 0x69
  db 0x91, 0x75, 0x51, 0x95, 0xaf, 0x73, 0x8b, 0xaf
  db 0x99, 0x5e, 0x4b, 0x7a, 0x74, 0x66, 0xa3, 0xb8
  db 0x7f, 0x5b, 0x81, 0x76, 0x57, 0x8a, 0xb4, 0x8d
  db 0x97, 0xa2, 0x94, 0x6c, 0x44, 0x75, 0x7f, 0x62
  db 0x83, 0xa7, 0x81, 0x67, 0x7c, 0x7e, 0x6b, 0x81
  db 0x95, 0x86, 0x89, 0x98, 0xa4, 0x86, 0x5f, 0x77
  db 0x83, 0x6d, 0x60, 0x83, 0x87, 0x72, 0x7c, 0x86
  db 0x7e, 0x83, 0x8b, 0x88, 0x87, 0x8d, 0x93, 0x9b
  db 0x88, 0x6c, 0x7a, 0x86, 0x7b, 0x62, 0x76, 0x79
  db 0x75, 0x77, 0x86, 0x81, 0x7b, 0x87, 0x8a, 0x8a
  db 0x8d, 0x92, 0x91, 0x91, 0x78, 0x7c, 0x7f, 0x80
  db 0x6f, 0x69, 0x72, 0x74, 0x76, 0x79, 0x7d, 0x7e
  db 0x80, 0x83, 0x8b, 0x8d, 0x8a, 0x8f, 0x91, 0x90
  db 0x8c, 0x83, 0x83, 0x7c, 0x7e, 0x71, 0x6a, 0x6f
  db 0x73, 0x76, 0x77, 0x79, 0x7b, 0x7f, 0x86, 0x8d
  db 0x8b, 0x89, 0x8a, 0x8f, 0x92, 0x92, 0x8b, 0x83
  db 0x7c, 0x7c, 0x78, 0x71, 0x6c, 0x6d, 0x71, 0x6f
  db 0x74, 0x7c, 0x7c, 0x7c, 0x81, 0x89, 0x8d, 0x8f
  db 0x91, 0x8d, 0x8c, 0x8d, 0x91, 0x8c, 0x81, 0x7d
  db 0x7b, 0x74, 0x6d, 0x6e, 0x71, 0x70, 0x6f, 0x78
  db 0x7d, 0x7e, 0x81, 0x85, 0x89, 0x8c, 0x90, 0x90
  db 0x89, 0x8d, 0x91, 0x88, 0x83, 0x84, 0x7e, 0x7a
  db 0x7b, 0x7a, 0x72, 0x6f, 0x74, 0x74, 0x73, 0x79
  db 0x7f, 0x7c, 0x7e, 0x88, 0x8a, 0x86, 0x8a, 0x8b
  db 0x87, 0x86, 0x8a, 0x89, 0x83, 0x80, 0x83, 0x81
  db 0x7e, 0x7d, 0x7b, 0x77, 0x76, 0x77, 0x77, 0x75
  db 0x78, 0x79, 0x7a, 0x7e, 0x81, 0x84, 0x85, 0x89
  db 0x8a, 0x8a, 0x8b, 0x8a, 0x8a, 0x87, 0x85, 0x83
  db 0x84, 0x84, 0x81, 0x7c, 0x7b, 0x7a, 0x77, 0x77
  db 0x76, 0x74, 0x74, 0x77, 0x79, 0x7d, 0x81, 0x81
  db 0x84, 0x87, 0x8a, 0x8c, 0x8c, 0x89, 0x86, 0x85
  db 0x85, 0x85, 0x83, 0x81, 0x7d, 0x7a, 0x7b, 0x7e
  db 0x7d, 0x7a, 0x77, 0x78, 0x77, 0x78, 0x79, 0x7c
  db 0x7b, 0x7a, 0x7f, 0x85, 0x86, 0x8a, 0x89, 0x87
  db 0x88, 0x88, 0x8a, 0x85, 0x82, 0x82, 0x7f, 0x7e
  db 0x7f, 0x81, 0x7e, 0x7b, 0x7e, 0x7e, 0x7b, 0x7a
  db 0x7a, 0x77, 0x73, 0x78, 0x7b, 0x78, 0x7a, 0x80
  db 0x81, 0x84, 0x89, 0x8c, 0x8b, 0x89, 0x8a, 0x88
  db 0x85, 0x88, 0x85, 0x80, 0x7f, 0x7f, 0x7d, 0x7e
  db 0x82, 0x80, 0x7d, 0x7b, 0x7a, 0x79, 0x78, 0x79
  db 0x75, 0x74, 0x76, 0x7a, 0x7d, 0x82, 0x84, 0x87
  db 0x8c, 0x8c, 0x90, 0x8d, 0x89, 0x85, 0x82, 0x80
  db 0x7d, 0x7e, 0x7c, 0x7a, 0x7a, 0x7f, 0x83, 0x80
  db 0x82, 0x84, 0x81, 0x7d, 0x7d, 0x7a, 0x73, 0x74
  db 0x78, 0x76, 0x74, 0x79, 0x83, 0x86, 0x89, 0x8f
  db 0x8a, 0x83, 0x85, 0x89, 0x87, 0x84, 0x85, 0x85
  db 0x84, 0x83, 0x8a, 0x89, 0x85, 0x87, 0x7a, 0x75
  db 0x74, 0x6f, 0x66, 0x63, 0x71, 0x7d, 0x7f, 0x82
  db 0x8b, 0x91, 0x95, 0x98, 0x97, 0x91, 0x88, 0x86
  db 0x7e, 0x78, 0x79, 0x75, 0x72, 0x71, 0x7e, 0x8e
  db 0xa1, 0x9e, 0x89, 0x7c, 0x6e, 0x5a, 0x46, 0x3c
  db 0x45, 0x5a, 0x6b, 0x7d, 0x87, 0x8e, 0x92, 0xa1
  db 0xaa, 0xa0, 0xa4, 0x9b, 0x93, 0x92, 0x95, 0x9a
  db 0x9f, 0x98, 0xa9, 0xbe, 0x9b, 0x85, 0x83, 0x77
  db 0x60, 0x59, 0x63, 0x5e, 0x5a, 0x5f, 0x6a, 0x73
  db 0x79, 0x86, 0x80, 0x68, 0x67, 0x61, 0x5b, 0x5f
  db 0x7f, 0xab, 0xb7, 0xc2, 0xb0, 0x99, 0x8b, 0x82
  db 0x79, 0x72, 0x6c, 0x6a, 0x60, 0x41, 0x40, 0x58
  db 0x5a, 0x5a, 0x6e, 0x63, 0x5b, 0x80, 0x92, 0x8d
  db 0x93, 0xb1, 0xb8, 0xd1, 0xe8, 0xd2, 0xb9, 0x94
  db 0x9c, 0x94, 0x86, 0x7e, 0x73, 0x6f, 0x64, 0x61
  db 0x67, 0x32, 0x37, 0x42, 0x2f, 0x4d, 0x84, 0x92
  db 0x7a, 0xaa, 0xd2, 0xcf, 0xc6, 0xb1, 0x58, 0x37
  db 0x83, 0x92, 0x8e, 0xc0, 0xb2, 0x95, 0x5b, 0x6e
  db 0x78, 0x44, 0x4d, 0x65, 0x7e, 0x9b, 0xb5, 0xa7
  db 0x8c, 0x7f, 0xa2, 0xa6, 0x96, 0x49, 0x21, 0x3f
  db 0x4f, 0x67, 0xc3, 0xda, 0xb1, 0x6b, 0x55, 0x7b
  db 0x59, 0x62, 0x81, 0x7c, 0x7f, 0xac, 0xae, 0x9a
  db 0xa4, 0xa3, 0x95, 0x8d, 0x4d, 0x15, 0x39, 0x48
  db 0x64, 0xb5, 0xc9, 0x9d, 0x4c, 0x4c, 0x66, 0x69
  db 0x71, 0x84, 0xcd, 0xcd, 0xaf, 0xa0, 0xa6, 0x9b
  db 0x8a, 0x8d, 0x4b, 0x15, 0x31, 0x4f, 0x6b, 0xa8
  db 0xd5, 0xa7, 0x4f, 0x42, 0x63, 0x76, 0x77, 0x8d
  db 0xbe, 0xc1, 0x8c, 0x7f, 0x7d, 0x82, 0x9f, 0xa6
  db 0x8a, 0x36, 0x36, 0x6b, 0x6e, 0x74, 0xba, 0xbc
  db 0x79, 0x41, 0x5a, 0x95, 0x82, 0x9c, 0x84, 0x63
  db 0x63, 0x9f, 0xca, 0xb9, 0xa2, 0x95, 0x79, 0x2e
  db 0x1e, 0x61, 0xa5, 0x93, 0xb2, 0x9f, 0x52, 0x35
  db 0x61, 0xb9, 0x88, 0x7e, 0xb6, 0x84, 0x3f, 0x74
  db 0xb6, 0x94, 0x8f, 0xb4, 0x7a, 0x38, 0x36, 0x8a
  db 0xb5, 0x76, 0xb1, 0xa7, 0x4b, 0x35, 0x7d, 0xc5
  db 0x9c, 0x70, 0x64, 0x59, 0x61, 0xb6, 0xc9, 0xa2
  db 0x8b, 0xa3, 0x82, 0x3c, 0x36, 0x83, 0xb2, 0x67
  db 0x9f, 0xb1, 0x64, 0x3b, 0x69, 0xb8, 0xa1, 0x54
  db 0x56, 0x66, 0x6c, 0xb7, 0xce, 0xb1, 0xa0, 0x9e
  db 0x4c, 0x2e, 0x58, 0xaa, 0xb0, 0x8e, 0xa1, 0x53
  db 0x31, 0x4f, 0x9e, 0xb8, 0x61, 0x61, 0x8c, 0x5b
  db 0x75, 0xbb, 0xbb, 0xa0, 0x9d, 0x69, 0x36, 0x4b
  db 0x81, 0xc3, 0x95, 0x8d, 0x6d, 0x43, 0x53, 0x8b
  db 0xc1, 0x92, 0x51, 0x57, 0x7b, 0x96, 0xb0, 0xbd
  db 0xa4, 0x93, 0x8b, 0x42, 0x38, 0x61, 0xb5, 0x96
  db 0x7c, 0x8e, 0x64, 0x4a, 0x61, 0xac, 0xb4, 0x65
  db 0x4d, 0x66, 0x76, 0xb6, 0xbe, 0xa8, 0xa2, 0x8e
  db 0x3f, 0x35, 0x65, 0xb3, 0xa1, 0x99, 0x8f, 0x3b
  db 0x46, 0x66, 0xba, 0xa7, 0x5c, 0x56, 0x66, 0x7b
  db 0xa3, 0xcc, 0xaa, 0xa7, 0x6f, 0x3e, 0x47, 0x87
  db 0xbc, 0xa4, 0x84, 0x47, 0x53, 0x58, 0x97, 0xc3
  db 0xa1, 0x5c, 0x53, 0x6e, 0x9f, 0xaf, 0xac, 0x91
  db 0x8e, 0x93, 0x39, 0x39, 0x63, 0xac, 0x86, 0x76
  db 0x87, 0x5f, 0x4d, 0x7c, 0xbf, 0xaf, 0x64, 0x57
  db 0x64, 0x95, 0xb8, 0xb2, 0xa5, 0x9d, 0x7f, 0x39
  db 0x40, 0x72, 0xba, 0x8a, 0x8e, 0x7e, 0x45, 0x4d
  db 0x77, 0xc3, 0x90, 0x52, 0x56, 0x75, 0x9f, 0xbd
  db 0xb3, 0xa1, 0x9a, 0x4c, 0x3b, 0x62, 0xb1, 0xae
  db 0x90, 0x86, 0x42, 0x44, 0x66, 0xb3, 0xaf, 0x61
  db 0x54, 0x6a, 0x88, 0xac, 0xb7, 0xa7, 0x78, 0x93
  db 0x6d, 0x3c, 0x58, 0xa3, 0xb1, 0x62, 0x64, 0x70
  db 0x63, 0x6f, 0xb3, 0xc2, 0x70, 0x51, 0x67, 0x94
  db 0xc4, 0xa4, 0xa0, 0x74, 0x8d, 0x7b, 0x3a, 0x39
  db 0x7f, 0xac, 0x72, 0x74, 0x7d, 0x69, 0x57, 0x8e
  db 0xc1, 0x91, 0x51, 0x62, 0x72, 0xac, 0xb7, 0xb2
  db 0x9b, 0x99, 0x6e, 0x33, 0x4b, 0x89, 0xb4, 0x7e
  db 0x93, 0x66, 0x48, 0x57, 0x8e, 0xc3, 0x7b, 0x55
  db 0x66, 0x79, 0xa3, 0xba, 0xb1, 0x93, 0x8c, 0x77
  db 0x3e, 0x4f, 0x90, 0xb6, 0x83, 0x4a, 0x5b, 0x64
  db 0x66, 0x9f, 0xca, 0x8d, 0x5c, 0x5d, 0x82, 0xbf
  db 0xaa, 0x94, 0x78, 0x66, 0x99, 0x8f, 0x47, 0x40
  db 0x7a, 0xa2, 0x6d, 0x81, 0xa1, 0x74, 0x55, 0x70
  db 0xb8, 0x95, 0x58, 0x5b, 0x6e, 0x99, 0xb7, 0xad
  db 0x9c, 0x8b, 0x9b, 0x69, 0x37, 0x4d, 0x93, 0x9f
  db 0x6b, 0x7e, 0x79, 0x53, 0x53, 0x99, 0xbe, 0x79
  db 0x50, 0x67, 0x82, 0xaa, 0xb1, 0xae, 0x97, 0x9b
  db 0x72, 0x32, 0x44, 0x80, 0xb9, 0x7b, 0x6c, 0x6b
  db 0x59, 0x60, 0x95, 0xce, 0x89, 0x54, 0x60, 0x86
  db 0xb6, 0xb9, 0xa4, 0x70, 0x4a, 0x8e, 0xa7, 0x62
  db 0x49, 0x79, 0xa7, 0x72, 0x6f, 0xac, 0x86, 0x54
  db 0x6d, 0xb6, 0x99, 0x58, 0x5e, 0x6b, 0x82, 0xb0
  db 0xb7, 0x8d, 0x6e, 0x89, 0xab, 0x7e, 0x43, 0x4c
  db 0x96, 0x99, 0x6f, 0x8f, 0x8b, 0x50, 0x49, 0x8a
  db 0xbb, 0x81, 0x54, 0x64, 0x68, 0x93, 0xb9, 0xae
  db 0x96, 0x92, 0x9c, 0x65, 0x35, 0x53, 0x9c, 0xa4
  db 0x6b, 0x7e, 0x73, 0x52, 0x61, 0xac, 0xc0, 0x74
  db 0x58, 0x69, 0x95, 0xad, 0xaa, 0xa8, 0x66, 0x46
  db 0x9a, 0xa8, 0x59, 0x49, 0x85, 0xa1, 0x5c, 0x73
  db 0xb7, 0x83, 0x4f, 0x73, 0xb6, 0x98, 0x5c, 0x63
  db 0x6b, 0x81, 0xac, 0xba, 0x89, 0x57, 0x80, 0xaf
  db 0x9f, 0x5e, 0x42, 0x72, 0x9d, 0x7a, 0x91, 0xa4
  db 0x62, 0x43, 0x68, 0xae, 0xa5, 0x62, 0x54, 0x5f
  db 0x6d, 0xac, 0xc2, 0x97, 0x73, 0x8b, 0xa8, 0x6e
  db 0x41, 0x52, 0x97, 0x97, 0x65, 0x95, 0x86, 0x54
  db 0x59, 0x9c, 0xbc, 0x70, 0x4f, 0x62, 0x7e, 0xa7
  db 0xb3, 0xaa, 0x77, 0x62, 0x89, 0xac, 0x7c, 0x3f
  db 0x5c, 0x93, 0x85, 0x6e, 0xa8, 0x88, 0x50, 0x5e
  db 0xa1, 0xb0, 0x77, 0x59, 0x60, 0x6e, 0xa3, 0xbd
  db 0xa3, 0x6d, 0x62, 0x94, 0xac, 0x90, 0x52, 0x4e
  db 0x76, 0x8f, 0x8f, 0xb0, 0x8e, 0x4b, 0x4c, 0x82
  db 0xb7, 0x97, 0x5c, 0x59, 0x5f, 0x91, 0xc0, 0xac
  db 0x81, 0x79, 0x97, 0x96, 0x60, 0x37, 0x59, 0x9d
  db 0x87, 0x7b, 0x9e, 0x6b, 0x4f, 0x6e, 0xb7, 0xa4
  db 0x60, 0x58, 0x75, 0xa4, 0xaf, 0xad, 0x97, 0x59
  db 0x62, 0xa9, 0x97, 0x4f, 0x45, 0x86, 0x99, 0x65
  db 0x8d, 0xa9, 0x66, 0x4f, 0x88, 0xbd, 0x8b, 0x5c
  db 0x61, 0x68, 0x88, 0xbb, 0xb6, 0x77, 0x53, 0x89
  db 0xac, 0xa7, 0x66, 0x44, 0x5d, 0x8c, 0x96, 0xa0
  db 0x9d, 0x59, 0x47, 0x6a, 0xa9, 0xad, 0x74, 0x4e
  db 0x58, 0x70, 0xb0, 0xb6, 0x90, 0x67, 0x7a, 0x99
  db 0xa1, 0x6b, 0x3c, 0x59, 0x8d, 0x93, 0x8a, 0xa4
  db 0x6c, 0x54, 0x64, 0xac, 0xad, 0x6c, 0x56, 0x6b
  db 0x8e, 0xae, 0xaf, 0x97, 0x68, 0x53, 0x8e, 0xb3
  db 0x94, 0x4b, 0x4d, 0x7e, 0x97, 0x88, 0xac, 0x8d
  db 0x4e, 0x51, 0x90, 0xb8, 0x89, 0x67, 0x54, 0x61
  db 0x84, 0xbe, 0xb1, 0x75, 0x4a, 0x7c, 0xa7, 0xab
  db 0x7e, 0x44, 0x55, 0x80, 0x9e, 0xa3, 0xa9, 0x6c
  db 0x44, 0x58, 0x97, 0xb8, 0x84, 0x4f, 0x54, 0x69
  db 0xa7, 0xbc, 0xa0, 0x6f, 0x65, 0x8a, 0xa0, 0x96
  db 0x4b, 0x45, 0x69, 0x9d, 0x89, 0x9b, 0x95, 0x59
  db 0x58, 0x7d, 0xbb, 0x99, 0x5e, 0x59, 0x70, 0x95
  db 0xb5, 0xaa, 0x7d, 0x52, 0x72, 0xa4, 0xaa, 0x6a
  db 0x47, 0x64, 0x97, 0x8e, 0x95, 0xab, 0x67, 0x48
  db 0x6c, 0xb1, 0xa4, 0x7d, 0x5c, 0x57, 0x64, 0x9f
  db 0xbb, 0x8f, 0x64, 0x62, 0x8c, 0xa6, 0xa4, 0x64
  db 0x41, 0x5d, 0x97, 0x9b, 0xa0, 0x9b, 0x5c, 0x47
  db 0x6b, 0xac, 0xa9, 0x69, 0x54, 0x66, 0x89, 0xb2
  db 0xab, 0x91, 0x64, 0x66, 0x92, 0xa9, 0x65, 0x3d
  db 0x6e, 0x95, 0x82, 0x84, 0xad, 0x78, 0x4d, 0x6d
  db 0xb2, 0xa0, 0x6d, 0x61, 0x62, 0x75, 0xa7, 0xbc
  db 0x8f, 0x5a, 0x6c, 0xa0, 0xb3, 0x7e, 0x4a, 0x58
  db 0x8c, 0x97, 0x8d, 0xad, 0x76, 0x4b, 0x58, 0x9f
  db 0xb0, 0x7e, 0x56, 0x56, 0x65, 0x9f, 0xbf, 0x98
  db 0x6d, 0x6b, 0x8f, 0xa3, 0x8b, 0x46, 0x4a, 0x7a
  db 0x98, 0x83, 0xa5, 0x86, 0x52, 0x55, 0x92, 0xbd
  db 0x85, 0x5c, 0x66, 0x7b, 0x9d, 0xb6, 0xa3, 0x74
  db 0x55, 0x86, 0xa9, 0x89, 0x45, 0x58, 0x8c, 0x89
  db 0x78, 0xa7, 0x91, 0x51, 0x57, 0xa6, 0xb1, 0x78
  db 0x6a, 0x5e, 0x61, 0x93, 0xbb, 0x9e, 0x63, 0x60
  db 0x95, 0xaf, 0x99, 0x56, 0x4b, 0x6c, 0x92, 0x92
  db 0xab, 0x8d, 0x57, 0x55, 0x83, 0xb4, 0x9a, 0x5f
  db 0x53, 0x65, 0x93, 0xb7, 0x9b, 0x83, 0x69, 0x83
  db 0x9e, 0x92, 0x48, 0x49, 0x77, 0x94, 0x7f, 0x96
  db 0x92, 0x57, 0x57, 0x8f, 0xba, 0x85, 0x5c, 0x65
  db 0x83, 0x96, 0xa7, 0xa1, 0x7a, 0x5a, 0x86, 0xb4
  db 0x99, 0x52, 0x52, 0x7f, 0x94, 0x7f, 0xa8, 0x93
  db 0x51, 0x50, 0x8d, 0xb4, 0x86, 0x6f, 0x54, 0x5a
  db 0x83, 0xbf, 0xa8, 0x79, 0x62, 0x79, 0x97, 0xac
  db 0x7b, 0x40, 0x55, 0x88, 0x98, 0x9b, 0xa8, 0x6f
  db 0x45, 0x5a, 0xa4, 0xae, 0x6e, 0x4e, 0x63, 0x86
  db 0xb3, 0xb4, 0x9e, 0x6f, 0x66, 0x90, 0xa8, 0x68
  db 0x3b, 0x66, 0x95, 0x88, 0x7d, 0xa8, 0x77, 0x4e
  db 0x6e, 0xad, 0x9f, 0x69, 0x63, 0x6f, 0x86, 0xaa
  db 0xae, 0x88, 0x60, 0x72, 0xa0, 0xad, 0x74, 0x45
  db 0x5a, 0x8f, 0x90, 0x91, 0xaf, 0x73, 0x4b, 0x60
  db 0xa4, 0xad, 0x79, 0x50, 0x56, 0x6f, 0xa8, 0xbb
  db 0x98, 0x72, 0x6e, 0x8f, 0xa3, 0x70, 0x3c, 0x53
  db 0x86, 0x8d, 0x85, 0xad, 0x74, 0x50, 0x61, 0xa4
  db 0xab, 0x62, 0x5e, 0x75, 0x8c, 0xa6, 0xb5, 0x97
  db 0x67, 0x71, 0xa2, 0x9a, 0x4f, 0x45, 0x7e, 0x91
  db 0x6e, 0x8e, 0xa6, 0x63, 0x51, 0x87, 0xb2, 0x87
  db 0x63, 0x62, 0x69, 0x8c, 0xb3, 0xab, 0x74, 0x63
  db 0x8f, 0xaf, 0x92, 0x50, 0x45, 0x76, 0x97, 0x87
  db 0xab, 0x95, 0x4d, 0x4c, 0x8b, 0xbc, 0x8e, 0x55
  db 0x54, 0x68, 0x99, 0xba, 0xa5, 0x7f, 0x7a, 0x90
  db 0x9b, 0x5f, 0x35, 0x6c, 0x9b, 0x86, 0x7b, 0xa6
  db 0x65, 0x44, 0x7a, 0xbe, 0x98, 0x52, 0x5e, 0x84
  db 0x9d, 0xa5, 0xb1, 0x7b, 0x53, 0x79, 0xb5, 0x95
  db 0x4e, 0x4d, 0x88, 0x90, 0x79, 0xab, 0x9a, 0x50
  db 0x51, 0x97, 0xaf, 0x7d, 0x62, 0x5b, 0x66, 0x9a
  db 0xc0, 0xa2, 0x62, 0x67, 0x9c, 0xac, 0x72, 0x43
  db 0x58, 0x8b, 0x86, 0x92, 0xb1, 0x77, 0x4b, 0x66
  db 0xa7, 0xa5, 0x6d, 0x56, 0x63, 0x80, 0xb0, 0xba
  db 0x90, 0x72, 0x7c, 0x98, 0x86, 0x3f, 0x50, 0x87
  db 0x88, 0x73, 0x8f, 0x92, 0x56, 0x5b, 0x95, 0xb1
  db 0x7d, 0x5e, 0x67, 0x80, 0xa2, 0xaf, 0x9b, 0x6b
  db 0x6c, 0x92, 0xad, 0x7f, 0x4b, 0x59, 0x8e, 0x8e
  db 0x82, 0xb2, 0x87, 0x4b, 0x5d, 0xa2, 0xab, 0x79
  db 0x5c, 0x55, 0x70, 0xa7, 0xbd, 0x94, 0x6e, 0x73
  db 0x91, 0xa1, 0x6a, 0x3e, 0x5c, 0x8d, 0x83, 0x8e
  db 0xab, 0x70, 0x4c, 0x68, 0xac, 0xa0, 0x69, 0x5d
  db 0x6e, 0x8d, 0xae, 0xb5, 0x92, 0x6b, 0x75, 0x9a
  db 0x99, 0x4e, 0x42, 0x83, 0x8e, 0x72, 0x8f, 0xa1
  db 0x5f, 0x48, 0x89, 0xb6, 0x88, 0x6d, 0x68, 0x64
  db 0x89, 0xb7, 0xa9, 0x6e, 0x63, 0x89, 0xae, 0xa5
  db 0x55, 0x41, 0x6d, 0x98, 0x87, 0xa2, 0x9d, 0x5a
  db 0x46, 0x83, 0xb6, 0x9a, 0x6a, 0x4f, 0x5d, 0x83
  db 0xbd, 0xaa, 0x80, 0x6e, 0x7f, 0x9c, 0x93, 0x4a
  db 0x44, 0x7a, 0x9d, 0x7a, 0x9a, 0x97, 0x52, 0x56
  db 0x8d, 0xbc, 0x80, 0x56, 0x60, 0x87, 0x9e, 0xb5
  db 0xa6, 0x77, 0x5c, 0x82, 0xad, 0x94, 0x49, 0x4d
  db 0x84, 0x88, 0x7f, 0xa6, 0x94, 0x53, 0x53, 0x91
  db 0xb0, 0x7f, 0x7a, 0x62, 0x55, 0x8c, 0xc1, 0xa3
  db 0x66, 0x69, 0x8e, 0xa4, 0xa2, 0x60, 0x40, 0x5e
  db 0x92, 0x96, 0xa4, 0x9a, 0x5e, 0x49, 0x73, 0xad
  db 0xa1, 0x70, 0x4f, 0x60, 0x88, 0xbd, 0xaa, 0x8a
  db 0x70, 0x77, 0x95, 0xa2, 0x5b, 0x39, 0x71, 0x8f
  db 0x7e, 0x87, 0xa1, 0x65, 0x4d, 0x84, 0xb3, 0x93
  db 0x71, 0x6a, 0x69, 0x85, 0xa9, 0xa6, 0x81, 0x63
  db 0x75, 0xa1, 0xaf, 0x62, 0x40, 0x67, 0x9b, 0x82
  db 0x96, 0xac, 0x64, 0x45, 0x75, 0xb9, 0x94, 0x7a
  db 0x5f, 0x50, 0x77, 0xb5, 0xb1, 0x80, 0x6c, 0x73
  db 0x91, 0xaa, 0x6f, 0x41, 0x55, 0x88, 0x88, 0x90
  db 0xb0, 0x72, 0x4d, 0x62, 0xa8, 0xa7, 0x7a, 0x5a
  db 0x64, 0x8c, 0xad, 0xae, 0x99, 0x73, 0x65, 0x87
  db 0xa7, 0x6e, 0x37, 0x6b, 0x90, 0x7b, 0x87, 0xac
  db 0x76, 0x48, 0x6e, 0xab, 0x97, 0x7c, 0x71, 0x5c
  db 0x75, 0xac, 0xb3, 0x84, 0x68, 0x7a, 0x9a, 0xae
  db 0x73, 0x41, 0x52, 0x88, 0x96, 0x96, 0xa9, 0x73
  db 0x49, 0x58, 0xa5, 0xab, 0x83, 0x54, 0x54, 0x76
  db 0xaf, 0xb5, 0x91, 0x7e, 0x78, 0x8a, 0x9c, 0x64
  db 0x33, 0x60, 0x8f, 0x90, 0x86, 0xa2, 0x71, 0x4c
  db 0x70, 0xac, 0x9f, 0x67, 0x59, 0x75, 0x91, 0xa4
  db 0xab, 0x8e, 0x61, 0x71, 0xa6, 0xa6, 0x60, 0x40
  db 0x76, 0x94, 0x81, 0x9b, 0xa9, 0x63, 0x4a, 0x7d
  db 0xb2, 0x8a, 0x74, 0x65, 0x52, 0x7e, 0xb4, 0xaf
  db 0x7a, 0x67, 0x80, 0xa1, 0xa0, 0x61, 0x3f, 0x61
  db 0x91, 0x89, 0xa7, 0xa9, 0x64, 0x4c, 0x7b, 0xa8
  db 0x93, 0x75, 0x5c, 0x66, 0x8a, 0xb3, 0xa9, 0x8c
  db 0x75, 0x78, 0x96, 0x97, 0x44, 0x41, 0x7e, 0x8e
  db 0x7d, 0x8d, 0x95, 0x5e, 0x57, 0x96, 0xaf, 0x85
  db 0x6b, 0x69, 0x6f, 0x90, 0xb0, 0x9e, 0x7a, 0x69
  db 0x82, 0xaa, 0xa0, 0x51, 0x43, 0x70, 0x9b, 0x86
  db 0xa4, 0x9b, 0x54, 0x4c, 0x83, 0xb1, 0x8f, 0x74
  db 0x56, 0x5f, 0x88, 0xba, 0xa8, 0x7f, 0x75, 0x7c
  db 0x97, 0x99, 0x50, 0x3d, 0x6a, 0x97, 0x7e, 0x9a
  db 0x9b, 0x57, 0x50, 0x83, 0xbb, 0x8b, 0x5e, 0x61
  db 0x7e, 0x9a, 0xb3, 0xac, 0x83, 0x66, 0x7d, 0xa4
  db 0x91, 0x43, 0x4d, 0x82, 0x8b, 0x7a, 0xa1, 0x93
  db 0x4d, 0x57, 0x98, 0xad, 0x7f, 0x6f, 0x61, 0x66
  db 0x9a, 0xba, 0x9a, 0x72, 0x6a, 0x89, 0xab, 0x96
  db 0x4f, 0x41, 0x71, 0x96, 0x8a, 0xad, 0x92, 0x54
  db 0x50, 0x89, 0xb3, 0x92, 0x62, 0x53, 0x67, 0x96
  db 0xbc, 0xa2, 0x80, 0x76, 0x8b, 0x9f, 0x7c, 0x3a
  db 0x4d, 0x83, 0x92, 0x7d, 0xa4, 0x81, 0x46, 0x5d
  db 0x9d, 0xb3, 0x73, 0x5e, 0x74, 0x8b, 0xa3, 0xb6
  db 0x9f, 0x69, 0x67, 0x91, 0xad, 0x7c, 0x3d, 0x5d
  db 0x86, 0x81, 0x89, 0xb5, 0x7c, 0x42, 0x61, 0xa8
  db 0xa3, 0x7b, 0x6d, 0x54, 0x73, 0xae, 0xb8, 0x89
  db 0x6b, 0x7a, 0x97, 0xab, 0x74, 0x3b, 0x50, 0x87
  db 0x8b, 0x97, 0xb1, 0x72, 0x42, 0x5f, 0xa5, 0xa9
  db 0x7c, 0x58, 0x5b, 0x81, 0xae, 0xb6, 0x99, 0x7d
  db 0x77, 0x95, 0x9a, 0x49, 0x3e, 0x76, 0x92, 0x7a
  db 0x88, 0x9e, 0x5b, 0x4f, 0x8b, 0xb3, 0x89, 0x67
  db 0x68, 0x74, 0x93, 0xb0, 0xa6, 0x7c, 0x69, 0x7e
  db 0xa8, 0x9c, 0x4f, 0x49, 0x74, 0x91, 0x7d, 0xa7
  db 0xa3, 0x59, 0x48, 0x86, 0xaf, 0x89, 0x7a, 0x5c
  db 0x5d, 0x8a, 0xbb, 0xa8, 0x7d, 0x77, 0x7f, 0x9c
  db 0x8a, 0x48, 0x41, 0x77, 0x8d, 0x79, 0xa8, 0x94
  db 0x56, 0x52, 0x90, 0xb1, 0x84, 0x64, 0x5e, 0x7a
  db 0xa4, 0xb0, 0xa3, 0x85, 0x76, 0x88, 0xa5, 0x72
  db 0x33, 0x64, 0x8e, 0x85, 0x76, 0xa4, 0x7a, 0x48
  db 0x70, 0xad, 0x9b, 0x70, 0x6d, 0x63, 0x77, 0xab
  db 0xb6, 0x8c, 0x69, 0x6f, 0x91, 0xad, 0x79, 0x3e
  db 0x54, 0x87, 0x89, 0x8d, 0xb5, 0x7c, 0x4d, 0x61
  db 0x9e, 0x9d, 0x86, 0x68, 0x57, 0x72, 0xa6, 0xb9
  db 0x92, 0x7f, 0x7c, 0x8c, 0x9d, 0x5e, 0x33, 0x58
  db 0x90, 0x88, 0x84, 0xa8, 0x6d, 0x4c, 0x6d, 0xaf
  db 0x9b, 0x6c, 0x61, 0x70, 0x8f, 0xaf, 0xb2, 0x8e
  db 0x6d, 0x6f, 0x99, 0xab, 0x5d, 0x38, 0x6f, 0x8d
  db 0x7d, 0x8f, 0xac, 0x69, 0x46, 0x7b, 0xae, 0x8d
  db 0x7b, 0x6f, 0x59, 0x79, 0xb6, 0xb1, 0x7e, 0x6f
  db 0x7f, 0xa0, 0xa9, 0x64, 0x3b, 0x5b, 0x8d, 0x8a
  db 0x9e, 0xab, 0x65, 0x46, 0x70, 0xad, 0x9b, 0x73
  db 0x59, 0x5e, 0x81, 0xb7, 0xb1, 0x8e, 0x7c, 0x7b
  db 0x99, 0x97, 0x47, 0x3c, 0x75, 0x8f, 0x7d, 0x8e
  db 0x9a, 0x57, 0x50, 0x90, 0xb3, 0x89, 0x61, 0x68
  db 0x7e, 0x98, 0xaa, 0xa2, 0x7e, 0x64, 0x7d, 0xad
  db 0x9d, 0x49, 0x46, 0x79, 0x8b, 0x7f, 0xa9, 0x9d
  db 0x50, 0x4b, 0x92, 0xac, 0x88, 0x7b, 0x5d, 0x5e
  db 0x93, 0xbd, 0x9e, 0x7f, 0x76, 0x7f, 0xa0, 0x8f
  db 0x48, 0x3d, 0x70, 0x8f, 0x83, 0xab, 0x9b, 0x55
  db 0x4b, 0x8c, 0xb2, 0x7e, 0x65, 0x6d, 0x7c, 0x95
  db 0xac, 0xa8, 0x86, 0x70, 0x85, 0xa4, 0x7b, 0x36
  db 0x57, 0x89, 0x7f, 0x78, 0xa6, 0x88, 0x49, 0x63
  db 0xa4, 0x9e, 0x74, 0x6b, 0x72, 0x80, 0x9f, 0xa9
  db 0x8c, 0x70, 0x75, 0x96, 0xb2, 0x7b, 0x40, 0x50
  db 0x87, 0x8c, 0x8d, 0xba, 0x7e, 0x48, 0x5d, 0xa4
  db 0xa0, 0x72, 0x65, 0x68, 0x79, 0xa5, 0xaf, 0x97
  db 0x84, 0x7d, 0x8b, 0x9e, 0x68, 0x33, 0x58, 0x8b
  db 0x88, 0x82, 0xad, 0x6d, 0x48, 0x67, 0xa7, 0x9d
  db 0x5b, 0x72, 0x8d, 0x8a, 0x95, 0xac, 0x95, 0x69
  db 0x7a, 0xa0, 0x9f, 0x5a, 0x3c, 0x73, 0x85, 0x79
  db 0x98, 0xaa, 0x62, 0x48, 0x81, 0xa9, 0x90, 0x72
  db 0x6b, 0x6f, 0x92, 0xa8, 0x9e, 0x82, 0x78, 0x86
  db 0xa7, 0x9b, 0x4f, 0x3b, 0x69, 0x92, 0x80, 0xa8
  db 0xa0, 0x54, 0x48, 0x7f, 0xaf, 0x7f, 0x6a, 0x76
  db 0x74, 0x88, 0xae, 0xaa, 0x84, 0x83, 0x92, 0x9d
  db 0x71, 0x29, 0x57, 0x87, 0x83, 0x83, 0xa5, 0x7f
  db 0x46, 0x6a, 0xa2, 0xa3, 0x6d, 0x69, 0x7f, 0x8a
  db 0x9a, 0xa8, 0x8e, 0x6d, 0x77, 0x9b, 0xac, 0x64
  db 0x35, 0x65, 0x8d, 0x7c, 0x95, 0xb6, 0x74, 0x45
  db 0x6a, 0xa7, 0x91, 0x69, 0x74, 0x76, 0x87, 0xa8
  db 0xac, 0x84, 0x7f, 0x8e, 0x9c, 0x8c, 0x3e, 0x37
  db 0x6d, 0x93, 0x7b, 0xa6, 0x99, 0x4c, 0x4b, 0x86
  db 0xac, 0x67, 0x6e, 0x93, 0x84, 0x83, 0xb3, 0xa7
  db 0x76, 0x86, 0x9e, 0x9e, 0x4f, 0x35, 0x76, 0x84
  db 0x7a, 0x99, 0xab, 0x58, 0x45, 0x7f, 0xa7, 0x81
  db 0x65, 0x80, 0x7b, 0x8a, 0xa8, 0xa7, 0x76, 0x7d
  db 0x9a, 0xac, 0x88, 0x3a, 0x41, 0x7c, 0x85, 0x8b
  db 0xb8, 0x8b, 0x48, 0x55, 0x94, 0x94, 0x63, 0x80
  db 0x87, 0x77, 0x91, 0xb4, 0x92, 0x81, 0x96, 0x9d
  db 0x8a, 0x36, 0x40, 0x76, 0x7f, 0x7f, 0xad, 0x90
  db 0x42, 0x53, 0x8d, 0xa6, 0x6b, 0x79, 0x8f, 0x7d
  db 0x8e, 0xb3, 0xa2, 0x7d, 0x8c, 0x93, 0xa1, 0x65
  db 0x35, 0x56, 0x80, 0x85, 0x93, 0xaa, 0x67, 0x46
  db 0x74, 0xa5, 0x90, 0x7a, 0x76, 0x6c, 0x84, 0xad
  db 0xab, 0x89, 0x7f, 0x84, 0x97, 0xa3, 0x68, 0x32
  db 0x57, 0x92, 0x80, 0x91, 0xb2, 0x69, 0x3a, 0x6c
  db 0xae, 0x85, 0x61, 0x7a, 0x7c, 0x7a, 0xa5, 0xb2
  db 0x91, 0x88, 0x8d, 0x91, 0x9c, 0x65, 0x2b, 0x52
  db 0x82, 0x83, 0x99, 0xb1, 0x6d, 0x45, 0x69, 0x9e
  db 0x96, 0x73, 0x72, 0x77, 0x81, 0xa1, 0xaa, 0x96
  db 0x83, 0x81, 0x8f, 0xa5, 0x85, 0x35, 0x3d, 0x77
  db 0x88, 0x89, 0xb1, 0x8f, 0x49, 0x52, 0x87, 0x99
  db 0x77, 0x7e, 0x80, 0x78, 0x92, 0xab, 0x9b, 0x91
  db 0x91, 0x8f, 0x96, 0x85, 0x3c, 0x3e, 0x80, 0x7d
  db 0x76, 0x94, 0x92, 0x5b, 0x5e, 0x8e, 0x90, 0x7a
  db 0x75, 0x6b, 0x84, 0xa0, 0xa2, 0x92, 0x8f, 0x8f
  db 0x91, 0x9a, 0x9f, 0x7c, 0x33, 0x3b, 0x68, 0x94
  db 0x95, 0xa9, 0x8e, 0x48, 0x49, 0x70, 0x89, 0x7a
  db 0x8a, 0x91, 0x88, 0x8b, 0xa2, 0xa8, 0xa1, 0x9d
  db 0x93, 0x84, 0x49, 0x33, 0x5e, 0x71, 0x83, 0x97
  db 0x8d, 0x6a, 0x67, 0x7a, 0x83, 0x8a, 0x87, 0x74
  db 0x7a, 0x95, 0x97, 0x95, 0x9b, 0x8f, 0x7a, 0x80
  db 0x96, 0x9c, 0x8e, 0x60, 0x46, 0x4e, 0x63, 0x7e
  db 0x8e, 0x8c, 0x7d, 0x76, 0x7c, 0x76, 0x64, 0x79
  db 0x96, 0x9b, 0x9a, 0x95, 0x9b, 0x9c, 0x9d, 0xa1
  db 0x8f, 0x89, 0x4e, 0x2f, 0x56, 0x6c, 0x81, 0x8f
  db 0x87, 0x6f, 0x72, 0x80, 0x7d, 0x79, 0x91, 0x9e
  db 0x96, 0x90, 0x93, 0x8e, 0x8a, 0x88, 0x82, 0x7b
  db 0x81, 0x80, 0x6a, 0x67, 0x73, 0x71, 0x6a, 0x73
  db 0x7a, 0x7a, 0x80, 0x88, 0x87, 0x7f, 0x74, 0x78
  db 0x8b, 0x9a, 0x98, 0x94, 0x90, 0x8f, 0x92, 0x94
  db 0x8e, 0x8c, 0x8e, 0x74, 0x59, 0x52, 0x55, 0x5e
  db 0x6a, 0x75, 0x74, 0x79, 0x7f, 0x71, 0x72, 0x92
  db 0xa2, 0xa0, 0xa1, 0x9f, 0x9f, 0xa0, 0xa0, 0x96
  db 0x88, 0x81, 0x73, 0x5d, 0x54, 0x57, 0x5a, 0x61
  db 0x6c, 0x76, 0x7c, 0x7e, 0x84, 0x88, 0x85, 0x93
  db 0x97, 0x90, 0x94, 0x96, 0x8e, 0x84, 0x8a, 0x87
  db 0x82, 0x83, 0x82, 0x74, 0x73, 0x7c, 0x71, 0x6f
  db 0x6d, 0x68, 0x6b, 0x74, 0x7a, 0x7a, 0x79, 0x85
  db 0x8c, 0x8b, 0x94, 0x9b, 0x97, 0x98, 0x94, 0x95
  db 0x9b, 0x8b, 0x87, 0x86, 0x81, 0x78, 0x62, 0x61
  db 0x64, 0x5b, 0x64, 0x69, 0x61, 0x6c, 0x79, 0x88
  db 0x98, 0x93, 0x9c, 0x9d, 0x99, 0x9f, 0x95, 0x90
  db 0x90, 0x84, 0x82, 0x7b, 0x69, 0x6f, 0x76, 0x69
  db 0x5c, 0x6a, 0x6f, 0x61, 0x6a, 0x75, 0x77, 0x84
  db 0x86, 0x8c, 0x97, 0x9b, 0xa3, 0x9f, 0xa3, 0xa3
  db 0x8f, 0x8d, 0x87, 0x7c, 0x83, 0x7e, 0x70, 0x71
  db 0x7f, 0x7d, 0x6b, 0x69, 0x70, 0x6f, 0x74, 0x74
  db 0x7a, 0x79, 0x7d, 0x87, 0x87, 0x92, 0x91, 0x89
  db 0x94, 0x8b, 0x83, 0x85, 0x7b, 0x7d, 0x7d, 0x83
  db 0x80, 0x6b, 0x73, 0x78, 0x6f, 0x6b, 0x63, 0x70
  db 0x71, 0x6b, 0x7f, 0x87, 0x88, 0x8e, 0x8f, 0x8e
  db 0x8f, 0x8e, 0x8f, 0x8b, 0x84, 0x79, 0x7d, 0x89
  db 0x7d, 0x77, 0x80, 0x82, 0x84, 0x87, 0x82, 0x87
  db 0x8c, 0x83, 0x83, 0x86, 0x8e, 0x8d, 0x88, 0x90
  db 0x8e, 0x88, 0x89, 0x82, 0x7e, 0x7f, 0x7a, 0x7c
  db 0x78, 0x72, 0x74, 0x6d, 0x78, 0x7d, 0x6e, 0x6d
  db 0x76, 0x7b, 0x74, 0x72, 0x74, 0x6e, 0x69, 0x6b
  db 0x70, 0x77, 0x78, 0x75, 0x7a, 0x7e, 0x86, 0x89
  db 0x88, 0x8c, 0x8f, 0x93, 0x8d, 0x93, 0x9a, 0x94
  db 0x98, 0x9a, 0x9c, 0x98, 0x8b, 0x85, 0x87, 0x86
  db 0x7f, 0x7b, 0x83, 0x82, 0x7e, 0x83, 0x7f, 0x80
  db 0x7f, 0x7c, 0x82, 0x86, 0x7f, 0x7a, 0x7c, 0x78
  db 0x72, 0x6f, 0x6a, 0x65, 0x64, 0x63, 0x6b, 0x73
  db 0x72, 0x72, 0x76, 0x7d, 0x7d, 0x7e, 0x83, 0x84
  db 0x89, 0x8f, 0x92, 0x96, 0x93, 0x8e, 0x8f, 0x89
  db 0x84, 0x81, 0x7d, 0x78, 0x7a, 0x7b, 0x79, 0x7a
  db 0x7d, 0x7d, 0x7e, 0x84, 0x7f, 0x80, 0x82, 0x83
  db 0x86, 0x8c, 0x8d, 0x89, 0x88, 0x86, 0x80, 0x82
  db 0x82, 0x80, 0x7e, 0x7e, 0x82, 0x83, 0x7f, 0x7f
  db 0x7f, 0x80, 0x83, 0x83, 0x83, 0x7f, 0x7f, 0x7c
  db 0x7e, 0x7e, 0x78, 0x7d, 0x7f, 0x7c, 0x7e, 0x82
  db 0x7b, 0x79, 0x7e, 0x7e, 0x7a, 0x7d, 0x79, 0x7a
  db 0x81, 0x80, 0x81, 0x87, 0x7f, 0x7e, 0x80, 0x79
  db 0x75, 0x75, 0x74, 0x75, 0x77, 0x7b, 0x80, 0x81
  db 0x83, 0x84, 0x85, 0x85, 0x86, 0x85, 0x83, 0x85
  db 0x84, 0x82, 0x84, 0x85, 0x81, 0x81, 0x7f, 0x7e
  db 0x7e, 0x7f, 0x82, 0x84, 0x84, 0x84, 0x86, 0x85
  db 0x88, 0x89, 0x87, 0x85, 0x82, 0x82, 0x81, 0x7d
  db 0x7c, 0x7f, 0x7b, 0x7b, 0x7e, 0x7f, 0x7e, 0x7c
  db 0x7d, 0x7f, 0x7e, 0x7b, 0x7f, 0x7e, 0x7c, 0x80
  db 0x81, 0x7e, 0x7c, 0x7e, 0x82, 0x80, 0x81, 0x84
  db 0x81, 0x7f, 0x81, 0x83, 0x85, 0x81, 0x80, 0x83
  db 0x85, 0x85, 0x82, 0x82, 0x81, 0x82, 0x83, 0x84
  db 0x81, 0x81, 0x82, 0x82, 0x82, 0x84, 0x84, 0x7e
  db 0x7d, 0x81, 0x80, 0x7e, 0x7c, 0x7a, 0x7d, 0x7c
  db 0x7c, 0x7c, 0x7c, 0x7a, 0x7d, 0x82, 0x80, 0x7c
  db 0x7f, 0x81, 0x7f, 0x81, 0x83, 0x80, 0x7d, 0x7f
  db 0x81, 0x80, 0x7e, 0x7b, 0x7c, 0x7d, 0x7f, 0x7f
  db 0x7d, 0x7c, 0x7b, 0x7d, 0x80, 0x80, 0x7b, 0x7d
  db 0x80, 0x7e, 0x82, 0x84, 0x81, 0x7f, 0x84, 0x86
  db 0x85, 0x86, 0x84, 0x82, 0x83, 0x86, 0x85, 0x81
  db 0x7f, 0x7f, 0x82, 0x83, 0x82, 0x7e, 0x7e, 0x7f
  db 0x80, 0x82, 0x83, 0x7d, 0x7a, 0x7d, 0x7f, 0x80
  db 0x80, 0x7f, 0x7e, 0x7f, 0x83, 0x83, 0x80, 0x80
  db 0x80, 0x81, 0x84, 0x80, 0x7e, 0x7f, 0x7f, 0x7f
  db 0x81, 0x81, 0x7d, 0x7b, 0x7e, 0x81, 0x80, 0x81
  db 0x80, 0x7e, 0x7d, 0x80, 0x82, 0x80, 0x7e, 0x80
  db 0x81, 0x80, 0x81, 0x81, 0x7f, 0x81, 0x84, 0x85
  db 0x84, 0x82, 0x81, 0x84, 0x84, 0x84, 0x84, 0x82
  db 0x80, 0x81, 0x83, 0x83, 0x7f, 0x7d, 0x7e, 0x7f
  db 0x7e, 0x7f, 0x7e, 0x7d, 0x7d, 0x7e, 0x81, 0x80
  db 0x7d, 0x7f, 0x80, 0x80, 0x80, 0x82, 0x7f, 0x7f
  db 0x7e, 0x81, 0x81, 0x7f, 0x7e, 0x7e, 0x7e, 0x7e
  db 0x7f, 0x7e, 0x7e, 0x7d, 0x80, 0x81, 0x81, 0x80
  db 0x80, 0x80, 0x81, 0x81, 0x82, 0x82, 0x7e, 0x7d
  db 0x81, 0x81, 0x7f, 0x80, 0x80, 0x7e, 0x7c, 0x7d
  db 0x7e, 0x7f, 0x7e, 0x7f, 0x81, 0x81, 0x80, 0x7f
  db 0x80, 0x80, 0x82, 0x81, 0x82, 0x7f, 0x7f, 0x82
  db 0x81, 0x80, 0x82, 0x82, 0x81, 0x80, 0x81, 0x82
  db 0x7f, 0x80, 0x80, 0x80, 0x80, 0x7e, 0x7e, 0x81
  db 0x80, 0x7f, 0x82, 0x80, 0x7f, 0x81, 0x82, 0x80
  db 0x80, 0x82, 0x83, 0x81, 0x82, 0x83, 0x82, 0x82
  db 0x81, 0x82, 0x82, 0x80, 0x7f, 0x81, 0x82, 0x80
  db 0x81, 0x81, 0x80, 0x80, 0x7f, 0x7f, 0x7d, 0x7d
  db 0x7f, 0x80, 0x7e, 0x7e, 0x7f, 0x7d, 0x7d, 0x7f
  db 0x80, 0x7d, 0x7d, 0x7e, 0x7f, 0x80, 0x7f, 0x80
  db 0x7e, 0x7e, 0x80, 0x80, 0x80, 0x7e, 0x7f, 0x81
  db 0x81, 0x81, 0x82, 0x81, 0x80, 0x80, 0x81, 0x81
  db 0x80, 0x80, 0x81, 0x81, 0x82, 0x81, 0x80, 0x7f
  db 0x7f, 0x80, 0x81, 0x80, 0x7f, 0x7e, 0x80, 0x80
  db 0x7f, 0x81

.align 32
  dc32 9
_string_67:
  db "mike kohn"
.align 16

.align 32
  dc32 12
_string_69:
  db "java grinder"
.align 16

  ; VDP register initialization (24 bytes)
.align 32
vdp_reg_init_table:
  dc.b  0x04  ; reg  0 = mode reg 1: no H interrupt
  dc.b  0x14  ; reg  1 = mode reg 2: blanked, no V interrupt, DMA enable
  dc.b  0x30  ; reg  2 = name table base for scroll A: 0xC000
  dc.b  0x3c  ; reg  3 = name table base for window:   0xF000
  dc.b  0x07  ; reg  4 = name table base for scroll B: 0xE000
  dc.b  0x6c  ; reg  5 = sprite attribute table base: 0xD800
  dc.b  0x00  ; reg  6 = unused register: 0x00
  dc.b  0x00  ; reg  7 = background color: 0x00
  dc.b  0x00  ; reg  8 = unused register: 0x00
  dc.b  0x00  ; reg  9 = unused register: 0x00
  dc.b  0xff  ; reg 10 = H interrupt register: 0xFF (esentially off)
  dc.b  0x00  ; reg 11 = mode reg 3: disable ext int, full H/V scroll
  dc.b  0x81  ; reg 12 = mode reg 4: 40 cell horiz mode, no interlace
  dc.b  0x37  ; reg 13 = H scroll table base: 0xFC00
  dc.b  0x00  ; reg 14 = unused register: 0x00
  dc.b  0x01  ; reg 15 = auto increment: 0x01
  dc.b  0x01  ; reg 16 = scroll size: V=32 cell, H=64 cell
  dc.b  0x00  ; reg 17 = window H position: 0x00
  dc.b  0x00  ; reg 18 = window V position: 0x00
  dc.b  0xff  ; reg 19 = DMA length count low:   0x00FF
  dc.b  0xff  ; reg 20 = DMA length count high:  0xFFxx
  dc.b  0x00  ; reg 21 = DMA source address low: 0xxxxx00
  dc.b  0x00  ; reg 22 = DMA source address mid: 0xxx00xx
  dc.b  0x80  ; reg 23 = DMA source address high: VRAM fill, addr = 0x00xxxx

  ; PSG initialization: set all channels to minimum volume
psg_reg_init_table:
  dc.b  0x9f,0xbf,0xdf,0xff

_print_string:
  move.l (-4,a2), d5
  eor.l d6, d6
_print_string_loop:
  move.b (a2)+, d6
  move.w d6, (a0)
  dbra d5, _print_string_loop
  rts

_load_fonts:
  move.w #((fontend - font) / 4) - 1, d6
  move.l #0x4c200000, (a1)  ; C00004 VRAM write to 0x0c20
  movea.l #font, a2         ; Point to font set
_load_fonts_loop:
  move.l (a2)+, (a0)        ; C00000 write next longword of charset to VDP
  dbra d6, _load_fonts_loop ; loop until done
  rts

.align 32
font:
  dc.l 0x01111100, 0x11000110, 0x11000110, 0x11000110 ; A
  dc.l 0x11111110, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x11111100, 0x11000110, 0x11000110, 0x11111100 ; B
  dc.l 0x11000110, 0x11000110, 0x11111100, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000110, 0x11000000 ; C
  dc.l 0x11000110, 0x11000110, 0x11111110, 0x00000000
  dc.l 0x11111100, 0x11000110, 0x11000110, 0x11000110 ; D
  dc.l 0x11000110, 0x11000110, 0x11111100, 0x00000000
  dc.l 0x11111110, 0x11000000, 0x11000000, 0x11111100 ; E
  dc.l 0x11000000, 0x11000000, 0x11111110, 0x00000000
  dc.l 0x11111110, 0x11000000, 0x11000000, 0x11111100 ; F
  dc.l 0x11000000, 0x11000000, 0x11000000, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000000, 0x11001110 ; G
  dc.l 0x11000110, 0x11000110, 0x11111110, 0x00000000
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x11111110 ; H
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x00111000, 0x00111000, 0x00111000, 0x00111000 ; I
  dc.l 0x00111000, 0x00111000, 0x00111000, 0x00000000
  dc.l 0x00000110, 0x00000110, 0x00000110, 0x00000110 ; J
  dc.l 0x00000110, 0x01100110, 0x01111110, 0x00000000
  dc.l 0x11000110, 0x11001100, 0x11111000, 0x11111000 ; K
  dc.l 0x11001100, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x01100000, 0x01100000, 0x01100000, 0x01100000 ; L
  dc.l 0x01100000, 0x01100000, 0x01111110, 0x00000000
  dc.l 0x11000110, 0x11101110, 0x11111110, 0x11010110 ; M
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x11000110, 0x11100110, 0x11110110, 0x11011110 ; N
  dc.l 0x11001110, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000110, 0x11000110 ; O
  dc.l 0x11000110, 0x11000110, 0x11111110, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000110, 0x11111110 ; P
  dc.l 0x11000000, 0x11000000, 0x11000000, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000110, 0x11000110 ; Q
  dc.l 0x11001110, 0x11001110, 0x11111110, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000110, 0x11111100 ; R
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x11111110, 0x11000110, 0x11000000, 0x11111110 ; S
  dc.l 0x00000110, 0x11000110, 0x11111110, 0x00000000
  dc.l 0x11111110, 0x00111000, 0x00111000, 0x00111000 ; T
  dc.l 0x00111000, 0x00111000, 0x00111000, 0x00000000
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x11000110 ; U
  dc.l 0x11000110, 0x11000110, 0x11111110, 0x00000000
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x11000110 ; V
  dc.l 0x01101100, 0x00111000, 0x00010000, 0x00000000
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x11010110 ; W
  dc.l 0x11111110, 0x11101110, 0x11000110, 0x00000000
  dc.l 0x11000110, 0x11000110, 0x11101110, 0x01111100 ; X
  dc.l 0x11101110, 0x11000110, 0x11000110, 0x00000000
  dc.l 0x11000110, 0x11000110, 0x11000110, 0x01101100 ; Y
  dc.l 0x00111000, 0x00111000, 0x00111000, 0x00000000
  dc.l 0x11111110, 0x00001110, 0x00011100, 0x00111000 ; Z
  dc.l 0x01110000, 0x11100000, 0x11111110, 0x00000000
  dc.l 0x00000000, 0x00000000, 0x00000000, 0x00000000 ; 7B = .
  dc.l 0x00000000, 0x01100000, 0x01100000, 0x00000000
  dc.l 0x02222200, 0x22000220, 0x22000000, 0x02222200 ; 7C = S
  dc.l 0x00000220, 0x22000220, 0x02222200, 0x00000000
  dc.l 0x02222220, 0x22000000, 0x22000000, 0x22222200 ; 7D = E
  dc.l 0x22000000, 0x22000000, 0x02222220, 0x00000000
  dc.l 0x02222200, 0x22000220, 0x22000000, 0x22002220 ; 7E = G
  dc.l 0x22000220, 0x22000220, 0x02222220, 0x00000000
  dc.l 0x00022000, 0x00222200, 0x00222200, 0x02200220 ; 7F = A
  dc.l 0x02200220, 0x22000022, 0x22022222, 0x00000000
fontend:

  ; Load Z80 with software up to 8k.  a3 = pointer to code
.align 32
_load_z80:
  move.w #0x100, (Z80_BUSREQ) ; Pause Z80
  move.w #0x100, (Z80_RESET)  ; Reset Z80
_load_z80_wait_reset:
  btst.b #0, (Z80_BUSREQ)
  bne.s _load_z80_wait_reset
  lea (Z80_RAM), a2          ; 8k RAM area
  move.l (-4,a3), d5         ; Code len
_load_z80_next_byte:
  move.b (a3)+, (a2)+
  dbf d5, _load_z80_next_byte
  move.w #0, (Z80_RESET)     ; Disable the Z80 reset.
  move.w #0, (Z80_BUSREQ)    ; Give the Z80 the bus back.
  move.w #0x100, (Z80_RESET) ; Reset Z80
  rts

_set_pattern_table:
  move.l #0x40000000, (a1)   ; C00004 VRAM write to 0x0000
  move.l (-4,a3), d5         ; Code len
_set_pattern_table_loop:
  move.l (a3)+, (a0)
  dbf d5, _set_pattern_table_loop
  rts

_set_image_data:
  move.l #0x40000003, d7     ; Set cursor position in VDP
  move.l d7, (a1)
  move.l (-4,a3), d5         ; Code len
  eor.w d6, d6
_set_image_data_loop:
  move.w (a3)+, (a0)
  add.w #1, d6
  cmp.w #40, d6
  bne.s _set_image_data_not_40
  eor.w d6, d6
  add.l #0x00800000, d7
  move.l d7, (a1)
_set_image_data_not_40:
  dbf d5, _set_image_data_loop
  rts

_set_palette_colors:
  move.l #0xc0000000, (a1)   ; C00004 write CRAM address 0x0000
  move.l (-4,a3), d5         ; Code len
_set_palette_colors_loop:
  move.w (a3)+, (a0)
  dbf d5, _set_palette_colors_loop
  rts

_init_bitmap:
  move.l d4, (-4,a7)         ; put d4 on stack (or really below)
  move.l #0x40000000, (a1)   ; C00004 VRAM write to 0x0000
  move.l #(40*28*16)+15, d5  ; Pattern len (plus 1)
  eor.l d7, d7
_clear_pattern_table_loop:
  move.w d7, (a0)
  dbf d5, _clear_pattern_table_loop
  move.l #0x40000003, d4     ; Set cursor position in VDP
  move.l d4, (a1)            ; C00004 VRAM write to 0xC000
  move.l #(40*28)-1, d5      ; data len
  eor.w d6, d6
_init_bitmap_loop:
  move.w d7, (a0)
  add.w #1, d6
  cmp.w #40, d6
  bne.s _init_bitmap_loop_not_40
  eor.w d6, d6
  add.l #0x00800000, d4
  move.l d4, (a1)
_init_bitmap_loop_not_40:
  addq.l #1, d7
  dbf d5, _init_bitmap_loop
  move.l (-4,a7), d4         ; recover d4
  move.l #0x60000003, d7     ; Set cursor position in VDP
  move.l d7, (a1)            ; C00004 VRAM write to 0xC000
  move.l #0x1000-1, d5       ; data len
  move.w #0x460, d7          ; blank pattern
_init_bitmap_other_loop:
  move.w d7, (a0)
  dbf d5, _init_bitmap_other_loop
  move.l (-4,a7), d4         ; recover d4
  rts

_clear_bitmap:
  move.l #0x40000000, (a1)   ; C00004 VRAM write to 0x0000
  move.l #(40*28*16)+15, d5  ; Pattern len (w*h*16 words) (plus 1)
  eor.l d6, d6
_clear_bitmap_loop:
  move.w d6, (a0)
  dbf d5, _clear_bitmap_loop
  rts

  ;; plot(x=d5, y=d6, d7=color);
_plot:
  move.w d4, (-8,a7) ; save register so it can be used for other stuff
  move.w d5, (-4,a7)
  move.w d6, (-6,a7)
  lsr.w #3, d5   ; block_x = x / 8
  and.w #0xfff8, d6 ; block_y = (block_y / 8) * 8
  add.l d6, d5   ; address = block_x + block_y
  lsl.l #2, d6   ; block_y *= 4
  add.l d6, d5   ; address = block_x + block_y
  lsl.l #5, d5   ; address *= 32
  move.w (-4,a7), d6
  and.l #0x7, d6 ; x = x % 8
  lsr.w #1, d6   ; x = x / 2
  add.l d6, d5   ; address += x
  move.w (-6,a7), d6
  and.l #0x7, d6 ; y = y % 8
  lsl.w #2, d6   ; y = y * 4
  add.l d6, d5   ; address += y
  move.l d5, d4  ; need the upper 2 bits
  rol.w #2, d4
  and.w #3, d4
  and.w #0x3ffe, d5; address &= 0xffff3ffe (write to even address)
  swap d5        ; address <<= 16
  or.b d4, d5    ; move upper two bits of address into lower
  move.l d5, (a1); read word from VRAM and save in temp space
  move.w (a0), d4
  or.l #0x40000000, d5
  move.l d5, (a1)
  move.w (-4,a7), d6
  and.w #0x3, d6 ; x = x & 0x3
  neg.w d6
  addq.w #3, d6
  lsl.w #2, d6
  lsl.w d6, d7
  move.w #0xfff0, d5
  rol.w d6, d5
  and.w d4, d5
  or.w d5, d7
  move.w d7, (a0)
  move.w (-8,a7), d4 ; restore register
  rts


